"""Model Registry REST API.

REST API for Model Registry to create and manage ML model metadata

The version of the OpenAPI document: v1alpha3
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
from typing import Any

from pydantic import BaseModel, ConfigDict, ValidationError, field_validator
from typing_extensions import Self

from mr_openapi.models.data_set import DataSet
from mr_openapi.models.doc_artifact import DocArtifact
from mr_openapi.models.metric import Metric
from mr_openapi.models.model_artifact import ModelArtifact
from mr_openapi.models.parameter import Parameter

ARTIFACT_ONE_OF_SCHEMAS = ["DataSet", "DocArtifact", "Metric", "ModelArtifact", "Parameter"]


class Artifact(BaseModel):
    """A metadata Artifact Entity."""

    # data type: ModelArtifact
    oneof_schema_1_validator: ModelArtifact | None = None
    # data type: DocArtifact
    oneof_schema_2_validator: DocArtifact | None = None
    # data type: DataSet
    oneof_schema_3_validator: DataSet | None = None
    # data type: Metric
    oneof_schema_4_validator: Metric | None = None
    # data type: Parameter
    oneof_schema_5_validator: Parameter | None = None
    actual_instance: DataSet | DocArtifact | Metric | ModelArtifact | Parameter | None = None
    one_of_schemas: set[str] = {"DataSet", "DocArtifact", "Metric", "ModelArtifact", "Parameter"}

    model_config = ConfigDict(
        validate_assignment=True,
        protected_namespaces=(),
    )

    discriminator_value_class_map: dict[str, str] = {}

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                msg = "If a position argument is used, only 1 is allowed to set `actual_instance`"
                raise ValueError(msg)
            if kwargs:
                msg = "If a position argument is used, keyword arguments cannot be used."
                raise ValueError(msg)
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(**kwargs)

    @field_validator("actual_instance")
    def actual_instance_must_validate_oneof(cls, v):
        Artifact.model_construct()
        error_messages = []
        match = 0
        # validate data type: ModelArtifact
        if not isinstance(v, ModelArtifact):
            error_messages.append(f"Error! Input type `{type(v)}` is not `ModelArtifact`")
        else:
            match += 1
        # validate data type: DocArtifact
        if not isinstance(v, DocArtifact):
            error_messages.append(f"Error! Input type `{type(v)}` is not `DocArtifact`")
        else:
            match += 1
        # validate data type: DataSet
        if not isinstance(v, DataSet):
            error_messages.append(f"Error! Input type `{type(v)}` is not `DataSet`")
        else:
            match += 1
        # validate data type: Metric
        if not isinstance(v, Metric):
            error_messages.append(f"Error! Input type `{type(v)}` is not `Metric`")
        else:
            match += 1
        # validate data type: Parameter
        if not isinstance(v, Parameter):
            error_messages.append(f"Error! Input type `{type(v)}` is not `Parameter`")
        else:
            match += 1
        if match > 1:
            # more than 1 match
            raise ValueError(
                "Multiple matches found when setting `actual_instance` in Artifact with oneOf schemas: DataSet, DocArtifact, Metric, ModelArtifact, Parameter. Details: "
                + ", ".join(error_messages)
            )
        if match == 0:
            # no match
            raise ValueError(
                "No match found when setting `actual_instance` in Artifact with oneOf schemas: DataSet, DocArtifact, Metric, ModelArtifact, Parameter. Details: "
                + ", ".join(error_messages)
            )
        return v

    @classmethod
    def from_dict(cls, obj: str | dict[str, Any]) -> Self:
        return cls.from_json(json.dumps(obj))

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Returns the object represented by the json string."""
        instance = cls.model_construct()
        error_messages = []
        match = 0

        # use oneOf discriminator to lookup the data type
        _data_type = json.loads(json_str).get("artifactType")
        if not _data_type:
            msg = "Failed to lookup data type from the field `artifactType` in the input."
            raise ValueError(msg)

        # check if data type is `DataSet`
        if _data_type == "dataset-artifact":
            instance.actual_instance = DataSet.from_json(json_str)
            return instance

        # check if data type is `DocArtifact`
        if _data_type == "doc-artifact":
            instance.actual_instance = DocArtifact.from_json(json_str)
            return instance

        # check if data type is `Metric`
        if _data_type == "metric":
            instance.actual_instance = Metric.from_json(json_str)
            return instance

        # check if data type is `ModelArtifact`
        if _data_type == "model-artifact":
            instance.actual_instance = ModelArtifact.from_json(json_str)
            return instance

        # check if data type is `Parameter`
        if _data_type == "parameter":
            instance.actual_instance = Parameter.from_json(json_str)
            return instance

        # deserialize data into ModelArtifact
        try:
            instance.actual_instance = ModelArtifact.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into DocArtifact
        try:
            instance.actual_instance = DocArtifact.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into DataSet
        try:
            instance.actual_instance = DataSet.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into Metric
        try:
            instance.actual_instance = Metric.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into Parameter
        try:
            instance.actual_instance = Parameter.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if match > 1:
            # more than 1 match
            raise ValueError(
                "Multiple matches found when deserializing the JSON string into Artifact with oneOf schemas: DataSet, DocArtifact, Metric, ModelArtifact, Parameter. Details: "
                + ", ".join(error_messages)
            )
        if match == 0:
            # no match
            raise ValueError(
                "No match found when deserializing the JSON string into Artifact with oneOf schemas: DataSet, DocArtifact, Metric, ModelArtifact, Parameter. Details: "
                + ", ".join(error_messages)
            )
        return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance."""
        if self.actual_instance is None:
            return "null"

        if hasattr(self.actual_instance, "to_json") and callable(self.actual_instance.to_json):
            return self.actual_instance.to_json()
        return json.dumps(self.actual_instance)

    def to_dict(self) -> dict[str, Any] | DataSet | DocArtifact | Metric | ModelArtifact | Parameter | None:
        """Returns the dict representation of the actual instance."""
        if self.actual_instance is None:
            return None

        if hasattr(self.actual_instance, "to_dict") and callable(self.actual_instance.to_dict):
            return self.actual_instance.to_dict()
        # primitive type
        return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance."""
        return pprint.pformat(self.model_dump())



"""Client for the model registry."""

from __future__ import annotations

from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import TypeVar, cast, overload
from urllib.parse import urlparse

from mr_openapi import (
    ApiClient,
    Configuration,
    ModelRegistryServiceApi,
)
from mr_openapi import (
    exceptions as mr_exceptions,
)

from ._utils import required_args
from .types import (
    Artifact,
    Experiment,
    ExperimentRun,
    ExperimentRunArtifact,
    ListOptions,
    ModelArtifact,
    ModelVersion,
    RegisteredModel,
)

ArtifactT = TypeVar("ArtifactT", bound=Artifact)


@dataclass
class ModelRegistryAPIClient:
    """Model registry API."""

    config: Configuration

    @classmethod
    def secure_connection(
        cls,
        server_address: str,
        port: int = 443,
        *,
        user_token: str,
        custom_ca: str | None = None,
    ) -> ModelRegistryAPIClient:
        """Constructor.

        Args:
            server_address: Server address.
            port: Server port. Defaults to 443.

        Keyword Args:
            user_token: The PEM-encoded user token as a string.
            custom_ca: The path to a PEM-
        """
        parsed_url = urlparse(server_address)
        host = server_address if parsed_url.port else f"{server_address}:{port}"

        return cls(
            Configuration(
                host=host,
                access_token=user_token,
                ssl_ca_cert=custom_ca,
            )
        )

    @classmethod
    def insecure_connection(
        cls,
        server_address: str,
        port: int,
        user_token: str | None = None,
    ) -> ModelRegistryAPIClient:
        """Constructor.

        Args:
            server_address: Server address.
            port: Server port.
            user_token: The PEM-encoded user token as a string.
        """
        parsed_url = urlparse(server_address)
        host = server_address if parsed_url.port else f"{server_address}:{port}"

        config = Configuration(
            host=host,
            access_token=user_token,
            verify_ssl=False,
        )
        return cls(config)

    @asynccontextmanager
    async def get_client(self) -> AsyncIterator[ModelRegistryServiceApi]:
        """Get a client for the model registry."""
        api_client = ApiClient(self.config)
        client = ModelRegistryServiceApi(api_client)

        try:
            yield client
        finally:
            await api_client.close()

    async def upsert_registered_model(self, registered_model: RegisteredModel) -> RegisteredModel:
        """Upsert a registered model.

        Updates or creates a registered model on the server.

        Args:
            registered_model: Registered model.

        Returns:
            New registered model.
        """
        async with self.get_client() as client:
            if registered_model.id:
                rm = await client.update_registered_model(registered_model.id, registered_model.update())
            else:
                rm = await client.create_registered_model(registered_model.create())

        return RegisteredModel.from_basemodel(rm)

    async def get_registered_model_by_id(self, id: str) -> RegisteredModel | None:
        """Fetch a registered model by its ID.

        Args:
            id: Registered model ID.

        Returns:
            Registered model.
        """
        async with self.get_client() as client:
            try:
                rm = await client.get_registered_model(id)
            except mr_exceptions.NotFoundException:
                return None

        return RegisteredModel.from_basemodel(rm)

    @overload
    async def get_registered_model_by_params(self, name: str): ...

    @overload
    async def get_registered_model_by_params(self, *, external_id: str): ...

    @required_args(("name",), ("external_id",))
    async def get_registered_model_by_params(
        self, name: str | None = None, external_id: str | None = None
    ) -> RegisteredModel | None:
        """Fetch a registered model by its name or external ID.

        Args:
            name: Registered model name.
            external_id: Registered model external ID.

        Returns:
            Registered model.
        """
        async with self.get_client() as client:
            try:
                rm = await client.find_registered_model(name=name, external_id=external_id)
            except mr_exceptions.NotFoundException:
                return None

        return RegisteredModel.from_basemodel(rm)

    async def get_registered_models(self, options: ListOptions | None = None) -> list[RegisteredModel]:
        """Fetch registered models.

        Args:
            options: Options for listing registered models.

        Returns:
            Registered models.
        """
        async with self.get_client() as client:
            rm_list = await client.get_registered_models(**(options or ListOptions()).as_options())

        if options:
            options.next_page_token = rm_list.next_page_token

        return [RegisteredModel.from_basemodel(rm) for rm in rm_list.items or []]

    async def upsert_model_version(
        self, model_version: ModelVersion, registered_model_id: str | None = None
    ) -> ModelVersion:
        """Upsert a model version.

        Updates or creates a model version on the server.

        Args:
            model_version: Model version to upsert.
            registered_model_id: ID of the registered model this version will be associated to. Can be None when updating an existing model version.

        Returns:
            New model version.
        """
        async with self.get_client() as client:
            if model_version.id:
                mv = await client.update_model_version(model_version.id, model_version.update())
            elif registered_model_id:
                mv = await client.create_model_version(model_version.create(registered_model_id=registered_model_id))
            else:
                msg = f"Registered model ID required for creating a new model version: {model_version}"
                raise ValueError(msg)

        return ModelVersion.from_basemodel(mv)

    async def get_model_version_by_id(self, model_version_id: str) -> ModelVersion | None:
        """Fetch a model version by its ID.

        Args:
            model_version_id: Model version ID.

        Returns:
            Model version.
        """
        async with self.get_client() as client:
            try:
                mv = await client.get_model_version(model_version_id)
            except mr_exceptions.NotFoundException:
                return None

        return ModelVersion.from_basemodel(mv)

    async def get_model_versions(
        self, registered_model_id: str, options: ListOptions | None = None
    ) -> list[ModelVersion]:
        """Fetch model versions by registered model ID.

        Args:
            registered_model_id: Registered model ID.
            options: Options for listing model versions.

        Returns:
            Model versions.
        """
        async with self.get_client() as client:
            mv_list = await client.get_registered_model_versions(
                registered_model_id, **(options or ListOptions()).as_options()
            )

        if options:
            options.next_page_token = mv_list.next_page_token

        return [ModelVersion.from_basemodel(mv) for mv in mv_list.items or []]

    @overload
    async def get_model_version_by_params(self, registered_model_id: str, name: str): ...

    @overload
    async def get_model_version_by_params(self, *, external_id: str): ...

    @required_args(
        (
            "registered_model_id",
            "name",
        ),
        ("external_id",),
    )
    async def get_model_version_by_params(
        self,
        registered_model_id: str | None = None,
        name: str | None = None,
        external_id: str | None = None,
    ) -> ModelVersion | None:
        """Fetch a model version by associated parameters.

        Either fetches by using external ID or by using registered model ID and version name.

        Args:
            registered_model_id: Registered model ID.
            name: Model version.
            external_id: Model version external ID.

        Returns:
            Model version.
        """
        async with self.get_client() as client:
            try:
                mv = await client.find_model_version(
                    name=name,
                    external_id=external_id,
                    parent_resource_id=registered_model_id,
                )
            except mr_exceptions.NotFoundException:
                return None

        return ModelVersion.from_basemodel(mv)

    async def upsert_model_artifact(self, model_artifact: ModelArtifact) -> ModelArtifact:
        """Upsert a model artifact.

        Updates or creates a model artifact on the server.

        Args:
            model_artifact: Model artifact to upsert.
            model_version_id: ID of the model version this artifact will be associated to.

        Returns:
            New model artifact.
        """
        async with self.get_client() as client:
            if not model_artifact.id:
                ma = await client.create_model_artifact(model_artifact.create())
            else:
                ma = await client.update_model_artifact(model_artifact.id, model_artifact.update())
        return ModelArtifact.from_basemodel(ma)

    async def upsert_model_version_artifact(self, artifact: ArtifactT, model_version_id: str) -> ArtifactT:
        """Creates a model version artifact.

        Creates a model version artifact on the server.

        Args:
            artifact: Model version artifact to upsert.
            model_version_id: ID of the model version this artifact will be associated to.

        Returns:
            New model version artifact.
        """
        async with self.get_client() as client:
            return cast(
                ArtifactT,
                Artifact.validate_artifact(
                    await client.upsert_model_version_artifact(model_version_id, artifact.wrap())
                ),
            )

    async def get_model_artifact_by_id(self, id: str) -> ModelArtifact | None:
        """Fetch a model artifact by its ID.

        Args:
            id: Model artifact ID.

        Returns:
            Model artifact.
        """
        async with self.get_client() as client:
            try:
                ma = await client.get_model_artifact(id)
            except mr_exceptions.NotFoundException:
                return None

        return ModelArtifact.from_basemodel(ma)

    @overload
    async def get_model_artifact_by_params(
        self,
        name: str,
        model_version_id: str,
    ): ...

    @overload
    async def get_model_artifact_by_params(self, *, external_id: str): ...

    @required_args(
        (
            "name",
            "model_version_id",
        ),
        ("external_id",),
    )
    async def get_model_artifact_by_params(
        self,
        name: str | None = None,
        model_version_id: str | None = None,
        external_id: str | None = None,
    ) -> ModelArtifact | None:
        """Fetch a model artifact either by external ID or by its name and the ID of its associated model version.

        Args:
            name: Model artifact name.
            model_version_id: ID of the associated model version.
            external_id: Model artifact external ID.

        Returns:
            Model artifact.
        """
        async with self.get_client() as client:
            try:
                ma = await client.find_model_artifact(
                    name=name,
                    parent_resource_id=model_version_id,
                    external_id=external_id,
                )
            except mr_exceptions.NotFoundException:
                return None

        return ModelArtifact.from_basemodel(ma)

    async def get_model_artifacts(
        self,
        model_version_id: str | None = None,
        options: ListOptions | None = None,
    ) -> list[ModelArtifact]:
        """Fetches model artifacts.

        Args:
            model_version_id: ID of the associated model version.
            options: Options for listing model artifacts.

        Returns:
            Model artifacts.
        """
        async with self.get_client() as client:
            if model_version_id:
                art_list = await client.get_model_version_artifacts(
                    model_version_id, **(options or ListOptions()).as_options()
                )
                if options:
                    options.next_page_token = art_list.next_page_token
                models = []
                for art in art_list.items or []:
                    converted = Artifact.validate_artifact(art)
                    if isinstance(converted, ModelArtifact):
                        models.append(converted)
                return models

            ma_list = await client.get_model_artifacts(**(options or ListOptions()).as_options())
            if options:
                options.next_page_token = ma_list.next_page_token
            return [ModelArtifact.from_basemodel(ma) for ma in ma_list.items or []]

    async def get_model_version_artifacts(
        self,
        model_version_id: str,
        options: ListOptions | None = None,
    ) -> list[Artifact]:
        """Fetches model artifacts.

        Args:
            model_version_id: ID of the associated model version.
            options: Options for listing model artifacts.

        Returns:
            Model artifacts.
        """
        async with self.get_client() as client:
            art_list = await client.get_model_version_artifacts(
                model_version_id, **(options or ListOptions()).as_options()
            )
        if options:
            options.next_page_token = art_list.next_page_token
        return [Artifact.validate_artifact(art) for art in art_list.items or []]

    async def upsert_experiment(self, experiment: Experiment) -> Experiment:
        """Upsert an experiment.

        Updates or creates an experiment on the server.

        Args:
            experiment: Experiment to upsert.
        """
        async with self.get_client() as client:
            if experiment.id:
                exp = await client.update_experiment(experiment.id, experiment.update())
            elif experiment.name:
                if exp := await self.get_experiment_by_name(experiment.name):  # type: ignore[assignment]
                    exp = await client.update_experiment(exp.id, experiment.update())  # type: ignore[arg-type]
                else:
                    exp = await client.create_experiment(experiment.create())
        return Experiment.from_basemodel(exp)

    async def get_experiment_by_name(self, name: str) -> Experiment | None:
        """Fetch an experiment by its name.

        Args:
            name: Experiment name.
        """
        async with self.get_client() as client:
            try:
                exp = await client.find_experiment(name=name)
            except mr_exceptions.NotFoundException:
                return None

        return Experiment.from_basemodel(exp)

    async def get_experiment_by_id(self, id: str | int) -> Experiment | None:
        """Fetch an experiment by its ID.

        Args:
            id: Experiment ID.
        """
        async with self.get_client() as client:
            try:
                exp = await client.get_experiment(str(id))
            except mr_exceptions.NotFoundException:
                return None

        return RegisteredModel.from_basemodel(exp)  # type: ignore[return-value,arg-type]

    async def get_experiments(self, options: ListOptions | None = None) -> list[Experiment]:
        """Fetch experiments.

        Args:
            options: Options for listing experiments.
        """
        async with self.get_client() as client:
            exp_list = await client.get_experiments(**(options or ListOptions()).as_options())
        if options:
            options.next_page_token = exp_list.next_page_token
        return [Experiment.from_basemodel(exp) for exp in exp_list.items or []]

    async def upsert_experiment_run(self, experiment_run: ExperimentRun) -> ExperimentRun:
        """Upsert an experiment run.

        Updates or creates an experiment run on the server.

        Args:
            experiment_run: Experiment run to upsert.
        """
        async with self.get_client() as client:
            if experiment_run.id:
                exp_run = await client.create_experiment_run(experiment_run.id, experiment_run.update())  # type: ignore[arg-type]
            else:
                exp_run = await client.create_experiment_run(experiment_run.create())

        return ExperimentRun.from_basemodel(exp_run)

    async def get_experiment_runs_by_experiment_id(
        self, experiment_id: str | int, options: ListOptions | None = None
    ) -> list[ExperimentRun]:
        """Fetch experiment runs by experiment ID.

        Args:
            experiment_id: Experiment ID.
            options: Options for listing experiment runs.
        """
        async with self.get_client() as client:
            try:
                exp_runs = await client.get_experiment_experiment_runs(
                    str(experiment_id), **(options or ListOptions()).as_options()
                )
            except mr_exceptions.NotFoundException:
                return []

        if options:
            options.next_page_token = exp_runs.next_page_token

        return [ExperimentRun.from_basemodel(exp_run) for exp_run in exp_runs.items or []]

    async def get_experiment_runs_by_experiment_name(
        self, experiment_name: str, options: ListOptions | None = None
    ) -> list[ExperimentRun]:
        """Fetch experiment runs by experiment name.

        Args:
            experiment_name: Experiment run to upsert.
            options: Options for listing experiment runs.

        """
        async with self.get_client() as client:
            try:
                exp = await self.get_experiment_by_name(experiment_name)
                if not exp:
                    return []
                exp_runs = await client.get_experiment_experiment_runs(
                    str(exp.id), **(options or ListOptions()).as_options()
                )
            except mr_exceptions.NotFoundException:
                return []

        if options:
            options.next_page_token = exp_runs.next_page_token

        return [ExperimentRun.from_basemodel(exp_run) for exp_run in exp_runs.items or []]

    async def get_experiment_run_by_experiment_and_run_id(
        self,
        run_id: str | int,
        experiment_name: str | None = None,
        experiment_id: str | int | None = None,
    ) -> ExperimentRun:
        """Fetch experiment run by experiment name / ID and the run ID.

        Args:
            run_id: Run ID.
            experiment_name: Experiment name.
            experiment_id: Experiment ID.
            options: Options for listing experiment runs.

        Returns:
            Experiment run.
        """
        async with self.get_client() as client:
            try:
                if experiment_name:
                    exp = await self.get_experiment_by_name(experiment_name)
                elif experiment_id:
                    exp = await self.get_experiment_by_id(str(experiment_id))
                else:
                    msg = "Either experiment_name or experiment_id must be provided"
                    raise ValueError(msg)
                if not exp:
                    return None  # type: ignore[return-value]

                exp_run = await client.get_experiment_run(str(run_id))
            except mr_exceptions.NotFoundException:
                return None  # type: ignore[return-value]

        return ExperimentRun.from_basemodel(exp_run)

    async def get_experiment_run_by_experiment_and_run_name(
        self,
        run_name: str,
        experiment_name: str | None = None,
        experiment_id: str | int | None = None,
        options: ListOptions | None = None,
    ) -> ExperimentRun:
        """Fetch experiment runs by experiment name / ID and the run ID.

        Args:
            run_name: Run name.
            experiment_name: Experiment name.
            experiment_id: Experiment ID.
            options: Options for listing experiment runs.

        Returns:
            Experiment run.
        """
        async with self.get_client() as client:
            exp = None
            try:
                if experiment_name:
                    exp = await self.get_experiment_by_name(experiment_name)
                elif experiment_id:
                    exp = await self.get_experiment_by_id(str(experiment_id))

                if not exp:
                    return None  # type: ignore[return-value]

                exp_run = await client.get_experiment_run(exp.id)  # type: ignore[arg-type]
            except mr_exceptions.NotFoundException:
                return None  # type: ignore[return-value]

        return ExperimentRun.from_basemodel(exp_run)

    async def get_experiment_run_by_id(self, id: str) -> ExperimentRun:
        """Fetch an experiment run by its ID.

        Args:
            id: Experiment run ID.
        """
        async with self.get_client() as client:
            try:
                exp_run = await client.get_experiment_run(id)
            except mr_exceptions.NotFoundException:
                return None  # type: ignore[return-value]

        return ExperimentRun.from_basemodel(exp_run)

    async def upsert_experiment_run_artifact(
        self, experiment_run_id: str, artifact: ExperimentRunArtifact
    ) -> ExperimentRunArtifact:
        """Upsert an experiment run artifact (parameter, metric, or dataset).

        Updates or creates an experiment run on the server.

        Args:
            experiment_run_id: Experiment run ID.
            artifact: Artifact to upsert.
        """
        async with self.get_client() as client:
            return Artifact.validate_artifact(  # type: ignore[return-value]
                await client.upsert_experiment_run_artifact(
                    experimentrun_id=experiment_run_id, artifact=artifact.wrap()
                )
            )

    @overload
    async def get_artifacts_by_experiment_run_params(self, run_id: str | int, options: ListOptions | None = None): ...

    @overload
    async def get_artifacts_by_experiment_run_params(
        self,
        run_name: str,
        experiment_name: str | None = None,
        options: ListOptions | None = None,
    ): ...

    @overload
    async def get_artifacts_by_experiment_run_params(
        self,
        run_name: str,
        experiment_id: str | int | None = None,
        options: ListOptions | None = None,
    ): ...

    @required_args(("run_id",), ("run_name", "experiment_name"), ("run_name", "experiment_id"))  # type: ignore[misc]
    async def get_artifacts_by_experiment_run_params(
        self,
        run_id: str | int | None = None,
        run_name: str | None = None,
        experiment_name: str | None = None,
        experiment_id: str | int | None = None,
        *,
        options: ListOptions | None = None,
    ) -> ExperimentRunArtifact:
        """Fetch a log by experiment run ID and name.

        Args:
            run_id: Experiment run ID.
            run_name: Experiment run name.
            experiment_name: Experiment name.
            experiment_id: Experiment ID.

        Keyword Args:
            options: Options for listing experiment run artifacts.
        """
        async with self.get_client() as client:
            try:
                if not run_id and run_name:
                    if experiment_name:
                        exp_runs = await self.get_experiment_runs_by_experiment_name(
                            experiment_name=experiment_name,
                            options=ListOptions(limit=100),
                        )
                    elif experiment_id:
                        exp_runs = await self.get_experiment_runs_by_experiment_id(
                            experiment_id=experiment_id,
                            options=ListOptions(limit=100),
                        )
                    else:
                        msg = "Either experiment_name or experiment_id must be provided"
                        raise ValueError(msg)

                    run = next((r for r in exp_runs if r.name == run_name), None)
                    if not run:
                        print(
                            f"Could not find run {run_name} "
                            f"in experiment {experiment_name} within the first 100 runs. "
                            "Please narrow your search by run id."
                        )
                        return []  # type: ignore[return-value]
                    run_id = run.id

                logs = await client.get_experiment_run_artifacts(
                    str(run_id), **(options or ListOptions()).as_options()
                )
            except mr_exceptions.NotFoundException:
                return []  # type: ignore[return-value]

        if options:
            options.next_page_token = logs.next_page_token
        return [Artifact.validate_artifact(log) for log in logs.items or []]  # type: ignore[return-value]

    async def get_artifacts(
        self,
        filter_query: str | None = None,
        artifact_type: str | None = None,
        options: ListOptions | None = None,
    ) -> list[Artifact]:
        """Get artifacts with filtering capabilities.

        Args:
            filter_query: A SQL-like query string to filter the list of entities.
            artifact_type: Specifies the artifact type for listing artifacts.
            options: Options for listing artifacts.

        Returns:
            List of artifacts matching the filter criteria.
        """
        # Create options dict with additional filters
        options_dict = (options or ListOptions()).as_options()
        if filter_query is not None:
            options_dict["filter_query"] = filter_query
        if artifact_type is not None:
            options_dict["artifact_type"] = artifact_type

        async with self.get_client() as client:
            try:
                response = await client.get_artifacts(**options_dict)
            except mr_exceptions.NotFoundException:
                return []

        if options:
            options.next_page_token = response.next_page_token
        return [Artifact.validate_artifact(artifact) for artifact in response.items or []]



"""Model Registry REST API.

REST API for Model Registry to create and manage ML model metadata

The version of the OpenAPI document: v1alpha3
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
from enum import Enum

from typing_extensions import Self


class ExperimentRunStatus(str, Enum):
    """The state of the Experiment Run. The state transitions are   [SCHEDULED ->] RUNNING -> FINISHED | FAILED | KILLED RUNNING: Run has been initiated. SCHEDULED: Run is scheduled to run at a later time. FINISHED: Run has completed. FAILED: Run execution failed. KILLED: Run killed by user."""

    """
    allowed enum values
    """
    RUNNING = "RUNNING"
    SCHEDULED = "SCHEDULED"
    FINISHED = "FINISHED"
    FAILED = "FAILED"
    KILLED = "KILLED"

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of ExperimentRunStatus from a JSON string."""
        return cls(json.loads(json_str))



"""TrustManager for TUF trust configuration initialization and caching."""

import hashlib
import json
import logging
import os
import tempfile
import urllib.error
import urllib.request
from pathlib import Path
from urllib import parse

import platformdirs
from sigstore._internal.tuf import TrustUpdater

from .exceptions import InitializationError

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class TrustManager:
    """Manages TUF trust configuration initialization and caching.

    Handles downloading, validating, and caching root.json from TUF servers,
    with support for explicit bootstrap (like cosign initialize).
    """

    def __init__(self, cache_dir: str | os.PathLike[str] | None = None):
        """Initialize TrustManager.

        Args:
            cache_dir: Base cache directory (default: platformdirs user_data_dir)
                Each TUF URL gets URL-encoded subdirectory for isolation
        """
        if cache_dir:
            self.base_cache_dir = Path(cache_dir)
        else:
            self.base_cache_dir = Path(platformdirs.user_data_dir("model-registry")) / "signing"

    def get_tuf_cache_dir(self, tuf_url: str) -> Path:
        """Get cache directory for a specific TUF URL with URL-encoded isolation."""
        repo_base = parse.quote(tuf_url, safe="")
        return self.base_cache_dir / repo_base

    def get_root_json_path(self, tuf_url: str) -> Path:
        """Get cached root.json path for bootstrap."""
        return self.get_tuf_cache_dir(tuf_url) / "root.json"

    def get_trust_root_path(self, tuf_url: str) -> Path:
        """Get trusted_root.json cache path."""
        return self.get_tuf_cache_dir(tuf_url) / "root" / "targets" / "trusted_root.json"

    def get_trust_config_path(self, tuf_url: str) -> Path:
        """Get trust_config.json cache path."""
        return self.get_tuf_cache_dir(tuf_url) / "trust_config.json"

    def initialize(
        self,
        tuf_url: str,
        root_url: str | None = None,
        root_checksum: str | None = None,
        fulcio_url: str | None = None,
        rekor_url: str | None = None,
        tsa_url: str | None = None,
        oidc_issuer: str | None = None,
        force: bool = False,
    ) -> dict:
        """Download and cache trust configuration from TUF.

        Args:
            tuf_url: TUF repository URL
            root_url: Explicit root.json URL (like cosign --root)
            root_checksum: SHA256 checksum to validate root.json
            fulcio_url: Fulcio CA URL (optional - may be extracted from TUF)
            rekor_url: Rekor transparency log URL (optional)
            tsa_url: Timestamp authority URL (optional)
            oidc_issuer: OIDC issuer URL
            force: If True, overwrite existing cached config

        Returns:
            Trust configuration dict ready for signing/verification

        Raises:
            InitializationError: If initialization fails
            FileExistsError: If config exists and force=False
        """
        config_path = self.get_trust_config_path(tuf_url)
        if config_path.exists() and not force:
            msg = f"Trust configuration already exists at {config_path}. Use force=True to overwrite."
            raise FileExistsError(msg)

        try:
            # Download trusted root
            trusted_root = self._download_trusted_root(tuf_url, root_url=root_url, root_checksum=root_checksum)

            # Transform and create config
            trusted_root = self._transform_and_fix_trust_root(trusted_root)
            trust_config = self._create_trust_config(trusted_root, fulcio_url, rekor_url, tsa_url, oidc_issuer)

            # Cache results
            self._cache_trust_config(tuf_url, trust_config)
            return trust_config

        except Exception as e:
            if isinstance(e, (InitializationError, FileExistsError)):
                raise
            msg = f"Trust initialization failed: {e}"
            raise InitializationError(msg) from e

    def _try_direct_trust_updater(self, tuf_url: str) -> dict | None:
        """Try TrustUpdater without bootstrap (auto-detect mode only)."""
        try:
            updater = TrustUpdater(url=tuf_url, offline=False)
            trusted_root_path = updater.get_trusted_root_path()
            logger.info(f"Downloaded (TrustUpdater): {Path(trusted_root_path).name}")
            with open(trusted_root_path) as f:
                return json.load(f)
        except Exception:
            logger.info("Direct TrustUpdater failed, trying with bootstrap...")
            return None

    def _load_cached_root(self, cache_path: Path, checksum: str | None) -> bytes | None:
        """Load cached root.json and validate checksum if provided."""
        if not cache_path.exists():
            return None

        try:
            cached_data = cache_path.read_bytes()
            if checksum:
                cached_checksum = hashlib.sha256(cached_data).hexdigest()
                if cached_checksum.lower() == checksum.lower():
                    logger.info("Using cached root.json (checksum verified)")
                    return cached_data
                logger.info("Cached root.json checksum mismatch, re-downloading...")
                return None
            logger.info("Using cached root.json")
            return cached_data
        except Exception as e:
            logger.info(f"Could not use cached root.json: {e}, re-downloading...")
            return None

    def _download_root_json(self, urls: list[str]) -> bytes:
        """Download root.json from list of URLs."""
        for url in urls:
            try:
                logger.info(f"Downloading from: {url}")
                with urllib.request.urlopen(url) as response:  # noqa: S310
                    return response.read()
            except urllib.error.HTTPError:
                continue
        msg = f"Could not download root.json from: {urls}"
        raise InitializationError(msg)

    def _validate_checksum(self, data: bytes, expected_checksum: str | None):
        """Validate SHA256 checksum if provided."""
        if expected_checksum:
            computed_checksum = hashlib.sha256(data).hexdigest()
            if computed_checksum.lower() != expected_checksum.lower():
                msg = f"Root.json checksum mismatch!\n  Expected: {expected_checksum}\n  Got:      {computed_checksum}"
                raise InitializationError(msg)
            logger.info(f"Checksum verified: {computed_checksum}")
        else:
            logger.info("No checksum provided (skipping validation)")

    def _bootstrap_trust_updater(self, tuf_url: str, root_data: bytes) -> dict:
        """Bootstrap TrustUpdater with root.json and return trusted_root."""
        with tempfile.TemporaryDirectory() as tmpdir:
            bootstrap_root = Path(tmpdir) / "root.json"
            bootstrap_root.write_bytes(root_data)

            updater = TrustUpdater(url=tuf_url, offline=False, bootstrap_root=bootstrap_root)
            trusted_root_path = updater.get_trusted_root_path()
            logger.info(f"Initialized TUF client: {Path(trusted_root_path).name}")

            with open(trusted_root_path) as f:
                return json.load(f)

    def _download_trusted_root(
        self,
        tuf_url: str,
        root_url: str | None = None,
        root_checksum: str | None = None,
    ) -> dict:
        """Download trusted_root.json from TUF server.

        Supports explicit root.json URL (cosign-style) or auto-detection.

        Args:
            tuf_url: TUF repository URL
            root_url: Explicit URL to root.json (optional, auto-detect if not provided)
            root_checksum: SHA256 checksum to validate root.json (optional)

        Returns:
            Parsed trusted_root.json as dict

        Raises:
            InitializationError: If download or initialization fails
        """
        try:
            # Try direct TrustUpdater first (auto-detect mode only)
            if not root_url:
                logger.info(f"Downloading trusted root from TUF: {tuf_url}")
                result = self._try_direct_trust_updater(tuf_url)
                if result:
                    return result

            # Prepare URLs to try
            if root_url:
                logger.info(f"Downloading root.json from explicit URL: {root_url}")
                root_urls = [root_url]
            else:
                tuf_url_clean = tuf_url.rstrip("/")
                root_urls = [
                    f"{tuf_url_clean}/root.json",
                    f"{tuf_url_clean}/1.root.json",
                ]

            # Try cache first
            root_cache_path = self.get_root_json_path(tuf_url)
            root_data = self._load_cached_root(root_cache_path, root_checksum)

            # Download if not cached
            if not root_data:
                root_data = self._download_root_json(root_urls)
                self._validate_checksum(root_data, root_checksum)
                logger.info(f"Downloaded root.json ({len(root_data)} bytes)")

                # Cache for future use
                root_cache_path.parent.mkdir(parents=True, exist_ok=True)
                root_cache_path.write_bytes(root_data)
                logger.info(f"Cached to: {root_cache_path}")

            # Bootstrap TrustUpdater with root.json
            return self._bootstrap_trust_updater(tuf_url, root_data)

        except InitializationError:
            raise
        except Exception as e:
            msg = f"Failed to download trusted root: {e}"
            raise InitializationError(msg) from e

    def _transform_and_fix_trust_root(self, trusted_root: dict) -> dict:
        """Apply transformations to trusted root."""
        logger.info("Transforming checkpointKeyId to logId...")
        return self._transform_checkpoint_to_logid(trusted_root)

    @staticmethod
    def _transform_checkpoint_to_logid(obj):
        """Transform checkpointKeyId to logId recursively (TUF v0.1 → v0.2)."""
        if isinstance(obj, dict):
            if "checkpointKeyId" in obj:
                obj["logId"] = obj.pop("checkpointKeyId")
            for key, value in obj.items():
                obj[key] = TrustManager._transform_checkpoint_to_logid(value)
        elif isinstance(obj, list):
            for i, item in enumerate(obj):
                obj[i] = TrustManager._transform_checkpoint_to_logid(item)
        return obj

    def _create_trust_config(
        self,
        trusted_root: dict,
        fulcio_url: str | None = None,
        rekor_url: str | None = None,
        tsa_url: str | None = None,
        oidc_issuer: str | None = None,
    ) -> dict:
        """Create complete trust configuration from trusted root."""
        tsa_url_full = None
        if tsa_url:
            tsa_url_full = tsa_url.rstrip("/") + "/api/v1/timestamp"

        return {
            "mediaType": "application/vnd.dev.sigstore.clienttrustconfig.v0.1+json",
            "trustedRoot": trusted_root,
            "signingConfig": {
                "mediaType": "application/vnd.dev.sigstore.signingconfig.v0.2+json",
                "caUrls": (
                    [
                        {
                            "url": fulcio_url,
                            "majorApiVersion": 1,
                            "validFor": {"start": "2020-01-01T00:00:00Z"},
                            "operator": "",
                        }
                    ]
                    if fulcio_url
                    else []
                ),
                "oidcUrls": (
                    [
                        {
                            "url": oidc_issuer,
                            "majorApiVersion": 1,
                            "validFor": {"start": "2020-01-01T00:00:00Z"},
                            "operator": "",
                        }
                    ]
                    if oidc_issuer
                    else []
                ),
                "rekorTlogUrls": (
                    [
                        {
                            "url": rekor_url,
                            "majorApiVersion": 1,
                            "validFor": {"start": "2020-01-01T00:00:00Z"},
                            "operator": "",
                        }
                    ]
                    if rekor_url
                    else []
                ),
                "tsaUrls": (
                    [
                        {
                            "url": tsa_url_full,
                            "majorApiVersion": 1,
                            "validFor": {"start": "2020-01-01T00:00:00Z"},
                            "operator": "",
                        }
                    ]
                    if tsa_url_full
                    else []
                ),
                "rekorTlogConfig": {"selector": "ANY"},
                "tsaConfig": {"selector": "ANY"},
            },
        }

    def _cache_trust_config(self, tuf_url: str, trust_config: dict):
        """Cache trust configuration to disk."""
        trust_root_path = self.get_trust_root_path(tuf_url)
        trust_config_path = self.get_trust_config_path(tuf_url)

        # Create cache directories
        trust_root_path.parent.mkdir(parents=True, exist_ok=True)
        trust_config_path.parent.mkdir(parents=True, exist_ok=True)

        # Cache trusted root
        trusted_root = trust_config.get("trustedRoot", {})
        trust_root_path.write_text(json.dumps(trusted_root, indent=2))
        logger.info(f"Cached trusted root to: {trust_root_path}")

        # Cache complete trust config
        trust_config_path.write_text(json.dumps(trust_config, indent=2))
        logger.info(f"Cached trust config to: {trust_config_path}")



"""Model Registry REST API.

REST API for Model Registry to create and manage ML model metadata

The version of the OpenAPI document: v1alpha3
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, ClassVar

from pydantic import BaseModel, ConfigDict, Field, StrictFloat, StrictInt, StrictStr
from typing_extensions import Self

from mr_openapi.models.artifact_state import ArtifactState
from mr_openapi.models.metadata_value import MetadataValue


class MetricCreate(BaseModel):
    """A metric to be created."""  # noqa: E501

    custom_properties: dict[str, MetadataValue] | None = Field(
        default=None,
        description="User provided custom properties which are not defined by its type.",
        alias="customProperties",
    )
    description: StrictStr | None = Field(default=None, description="An optional description about the resource.")
    external_id: StrictStr | None = Field(
        default=None,
        description="The external id that come from the clients’ system. This field is optional. If set, it must be unique among all resources within a database instance.",
        alias="externalId",
    )
    name: StrictStr | None = Field(
        default=None, description='The name/key of the metric (e.g., "accuracy", "loss", "f1_score").'
    )
    artifact_type: StrictStr | None = Field(default="metric", alias="artifactType")
    value: StrictFloat | StrictInt | None = Field(default=None, description="The numeric value of the metric.")
    timestamp: StrictStr | None = Field(
        default=None, description="Unix timestamp in milliseconds when the metric was recorded."
    )
    step: StrictInt | None = Field(
        default=None, description="The step number for multi-step metrics (e.g., training epochs)."
    )
    state: ArtifactState | None = ArtifactState.UNKNOWN
    __properties: ClassVar[list[str]] = [
        "customProperties",
        "description",
        "externalId",
        "name",
        "artifactType",
        "value",
        "timestamp",
        "step",
        "state",
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self | None:
        """Create an instance of MetricCreate from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: set[str] = set()

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each value in custom_properties (dict)
        _field_dict = {}
        if self.custom_properties:
            for _key_custom_properties in self.custom_properties:
                if self.custom_properties[_key_custom_properties]:
                    _field_dict[_key_custom_properties] = self.custom_properties[_key_custom_properties].to_dict()
            _dict["customProperties"] = _field_dict
        return _dict

    @classmethod
    def from_dict(cls, obj: dict[str, Any] | None) -> Self | None:
        """Create an instance of MetricCreate from a dict."""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        return cls.model_validate(
            {
                "customProperties": {_k: MetadataValue.from_dict(_v) for _k, _v in obj["customProperties"].items()}
                if obj.get("customProperties") is not None
                else None,
                "description": obj.get("description"),
                "externalId": obj.get("externalId"),
                "name": obj.get("name"),
                "artifactType": obj.get("artifactType") if obj.get("artifactType") is not None else "metric",
                "value": obj.get("value"),
                "timestamp": obj.get("timestamp"),
                "step": obj.get("step"),
                "state": obj.get("state") if obj.get("state") is not None else ArtifactState.UNKNOWN,
            }
        )



"""Signing utilities for model registry."""

from model_registry.signing.exceptions import (
    BaseSigningError,
    InitializationError,
    SigningError,
    VerificationError,
)
from model_registry.signing.image_signer import CommandRunner, ImageSigner
from model_registry.signing.model_signer import ModelSigner
from model_registry.signing.token import decode_jwt_payload, extract_client_id
from model_registry.signing.trust_manager import TrustManager

__all__ = [
    "CommandRunner",
    "ImageSigner",
    "ModelSigner",
    "TrustManager",
    "BaseSigningError",
    "InitializationError",
    "SigningError",
    "VerificationError",
    "decode_jwt_payload",
    "extract_client_id",
]



"""Pager for iterating over items."""

from __future__ import annotations

import asyncio
from collections.abc import AsyncIterator, Awaitable, Callable, Iterator
from dataclasses import dataclass, field
from typing import Generic, TypeVar, cast

from .base import BaseModel
from .options import ListOptions, OrderByField

T = TypeVar("T", bound=BaseModel)


@dataclass
class Pager(Generic[T], Iterator[T], AsyncIterator[T]):
    """Pager for iterating over items.

    Assumes that page_fn is a paged function that takes ListOptions and returns a list of items.
    """

    page_fn: Callable[[ListOptions], list[T]] | Callable[[ListOptions], Awaitable[list[T]]]
    options: ListOptions = field(default_factory=ListOptions)

    def __post_init__(self):
        self.restart()
        if asyncio.iscoroutinefunction(self.page_fn):
            self.__next__ = NotImplemented  # type: ignore[method-assign]
            self.next_page = self._anext_page
            self.next_item = self._anext_item
        else:
            self.__anext__ = NotImplemented  # type: ignore[method-assign]
            self.next_page = self._next_page  # type: ignore[assignment]
            self.next_item = self._next_item  # type: ignore[assignment]

    def restart(self) -> Pager[T]:
        """Reset the pager.

        This keeps the current options and page function, but resets the internal state.
        """
        # as MLMD loops over pages, we need to keep track of the first page or we'll loop forever
        self._start: str | None = None
        self._current_page: list[T] | None = None
        # tracks the next item on the current page
        self._i = 0
        self.options.next_page_token = None
        return self

    def order_by_creation_time(self) -> Pager[T]:
        """Order items by creation time.

        This resets the pager.
        """
        self.options.order_by = OrderByField.CREATE_TIME
        return self.restart()

    def order_by_update_time(self) -> Pager[T]:
        """Order items by update time.

        This resets the pager.
        """
        self.options.order_by = OrderByField.LAST_UPDATE_TIME
        return self.restart()

    def order_by_id(self) -> Pager[T]:
        """Order items by ID.

        This resets the pager.
        """
        self.options.order_by = OrderByField.ID
        return self.restart()

    def page_size(self, n: int) -> Pager[T]:
        """Set the page size for each request.

        This resets the pager.
        """
        if n < 1:
            msg = f"Page size must be at least 1, got {n}"
            raise ValueError(msg)
        self.options.limit = n
        return self.restart()

    def ascending(self) -> Pager[T]:
        """Order items in ascending order.

        This resets the pager.
        """
        self.options.is_asc = True
        return self.restart()

    def descending(self) -> Pager[T]:
        """Order items in descending order.

        This resets the pager.
        """
        self.options.is_asc = False
        return self.restart()

    def _next_page(self) -> list[T]:
        """Get the next page of items.

        This will automatically loop over pages.
        """
        return cast(list[T], self.page_fn(self.options))

    async def _anext_page(self) -> list[T]:
        """Get the next page of items.

        This will automatically loop over pages.
        """
        return await cast(Awaitable[list[T]], self.page_fn(self.options))

    def _needs_fetch(self) -> bool:
        return not self._current_page or (self._i >= len(self._current_page) and self._start is not None)

    def _next_item(self) -> T:
        """Get the next item in the pager.

        This variant won't check for looping, so it's useful for manual iteration/scripting.

        NOTE: This won't check for looping, so use with caution.
        If you want to check for looping, use the pythonic `next()`.
        """
        if self._needs_fetch():
            self._current_page = self._next_page()
            self._i = 0
        if self._current_page is None:  # for example when the MR server is empty
            raise StopIteration
        if self._i >= len(self._current_page):
            raise StopIteration

        item = self._current_page[self._i]
        self._i += 1
        return item

    async def _anext_item(self) -> T:
        """Get the next item in the pager.

        This variant won't check for looping, so it's useful for manual iteration/scripting.

        NOTE: This won't check for looping, so use with caution.
        If you want to check for looping, use the pythonic `next()`.
        """
        if self._needs_fetch():
            self._current_page = await self._anext_page()
            self._i = 0
        if self._current_page is None:  # for example when the MR server is empty
            raise StopAsyncIteration
        if self._i >= len(self._current_page):
            raise StopAsyncIteration

        item = self._current_page[self._i]
        self._i += 1
        return item

    def __next__(self) -> T:
        check_looping = self._needs_fetch()

        item = self._next_item()

        if self._start is None:
            self._start = self.options.next_page_token
        elif check_looping and self.options.next_page_token == self._start:
            raise StopIteration

        return item

    async def __anext__(self) -> T:
        check_looping = self._needs_fetch()

        item = await self._anext_item()

        if self._start is None:
            self._start = self.options.next_page_token
        elif check_looping and self.options.next_page_token == self._start:
            raise StopAsyncIteration

        return item

    def __iter__(self) -> Iterator[T]:
        return self

    def __aiter__(self) -> AsyncIterator[T]:
        return self



"""Model Registry REST API.

REST API for Model Registry to create and manage ML model metadata

The version of the OpenAPI document: v1alpha3
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, ClassVar

from pydantic import BaseModel, ConfigDict, Field, StrictInt, StrictStr
from typing_extensions import Self

from mr_openapi.models.artifact import Artifact


class ArtifactList(BaseModel):
    """A list of Artifact entities."""  # noqa: E501

    next_page_token: StrictStr = Field(
        description="Token to use to retrieve next page of results.", alias="nextPageToken"
    )
    page_size: StrictInt = Field(description="Maximum number of resources to return in the result.", alias="pageSize")
    size: StrictInt = Field(description="Number of items in result list.")
    items: list[Artifact] = Field(description="Array of `Artifact` entities.")
    __properties: ClassVar[list[str]] = ["nextPageToken", "pageSize", "size", "items"]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self | None:
        """Create an instance of ArtifactList from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: set[str] = set()

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each item in items (list)
        _items = []
        if self.items:
            for _item_items in self.items:
                if _item_items:
                    _items.append(_item_items.to_dict())
            _dict["items"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: dict[str, Any] | None) -> Self | None:
        """Create an instance of ArtifactList from a dict."""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        return cls.model_validate(
            {
                "nextPageToken": obj.get("nextPageToken"),
                "pageSize": obj.get("pageSize"),
                "size": obj.get("size"),
                "items": [Artifact.from_dict(_item) for _item in obj["items"]]
                if obj.get("items") is not None
                else None,
            }
        )



"""Token utility functions for model signing."""

import base64
import json


def decode_jwt_payload(token: str) -> dict:
    """Decode JWT token payload to dictionary.

    Args:
        token: JWT token string in format header.payload.signature

    Returns:
        Dictionary of claims from the token payload

    Raises:
        ValueError: If token format is invalid or decoding fails

    Note:
        JWT format per OpenID spec: "A JWT is represented as the concatenation of
        the Encoded JWT Header, the JWT Second Part, and the JWT Third Part, in that
        order, with the parts being separated by period ('.') characters."
        https://openid.net/specs/draft-jones-json-web-token-07.html
    """
    try:
        # Validate JWT has exactly 3 parts (header.payload.signature)
        parts = token.split(".")
        if len(parts) != 3:
            msg = f"Invalid JWT format: expected 3 parts separated by '.', got {len(parts)}"
            raise ValueError(msg)

        payload = parts[1]
        # Add proper padding for base64 decoding
        padding = len(payload) % 4
        if padding:
            payload += "=" * (4 - padding)
        return json.loads(base64.b64decode(payload))
    except (IndexError, json.JSONDecodeError, Exception) as e:
        msg = f"Failed to decode JWT token payload: {e}"
        raise ValueError(msg) from e


def extract_client_id(claims: dict) -> str | None:
    """Extract client_id from decoded JWT claims' aud (audience) field.

    The aud claim can be either a string or an array of strings.
    If it's an array, returns the first element.

    Args:
        claims: Decoded JWT claims dictionary

    Returns:
        Client ID from aud claim, or None if not present or empty
    """
    aud = claims.get("aud")

    if isinstance(aud, list):
        return aud[0] if aud else None
    return aud



from __future__ import annotations  # noqa: I001

import time
from contextlib import AbstractContextManager
from dataclasses import dataclass
from typing import Any, Literal
from collections.abc import Callable
from types import TracebackType

from model_registry.core import ModelRegistryAPIClient
from model_registry.exceptions import StoreError
from model_registry.types.artifacts import (
    ArtifactState,
    DataSet,
    ExperimentRunArtifact,
    ExperimentRunArtifactTypes,
    Metric,
    Parameter,
    ParameterType,
)
from model_registry.types.experiments import ExperimentRun

from .utils import S3Params, ThreadSafeVariable, upload_to_s3

LogType = Literal["params", "metrics", "datasets"]


@dataclass
class RunContext:
    id: str | None = None
    name: str | None = None
    run_id: str | None = None
    active: bool = False


@dataclass
class RunInfo:
    experiment_id: str
    id: str
    name: str


class ActiveExperimentRun(AbstractContextManager):
    def __init__(
        self,
        thread_safe_ctx: ThreadSafeVariable,
        experiment_run: ExperimentRun,
        api: ModelRegistryAPIClient,
        async_runner: Callable,
    ):
        self._thread_safe_ctx = thread_safe_ctx
        self._exp_run = experiment_run
        self.info = RunInfo(
            id=experiment_run.id,  # type: ignore[arg-type]
            name=experiment_run.name,
            experiment_id=experiment_run.experiment_id,
        )
        self.__api = api
        self.__async_runner = async_runner
        self._logs: ExperimentRunArtifactTypes = ExperimentRunArtifactTypes()

    def __enter__(self) -> ActiveExperimentRun:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: TracebackType | None,
    ) -> None:
        """Exit the context manager and upsert the logs to the experiment run."""
        temp_artifacts: ExperimentRunArtifactTypes = ExperimentRunArtifactTypes()
        for log in self.get_logs():
            server_log = self.__async_runner(
                self.__api.upsert_experiment_run_artifact(experiment_run_id=self.info.id, artifact=log)
            )
            log_type = type(server_log)
            if log_type is Parameter:
                temp_artifacts.params[log.name] = server_log  # type: ignore[index]
            elif log_type is Metric:
                temp_artifacts.metrics[log.name] = server_log  # type: ignore[index]
            elif log_type is DataSet:
                temp_artifacts.datasets[log.name] = server_log  # type: ignore[index]
        self._logs = temp_artifacts
        self._thread_safe_ctx.set(RunContext(active=False))

    def log_param(self, key: str, value: Any, *, description: str | None = None):
        """Log a parameter to the experiment run.

        The parameter type is inferred from the value type.

        Args:
            key: Name of the parameter.
            value: Value of the parameter.

        Keyword Args:
            description: Description of the parameter.
        """
        param_type = ParameterType.STRING  # consistent with param_type default init
        if isinstance(value, bool):
            param_type = ParameterType.BOOLEAN
        elif isinstance(value, int | float):
            # TODO: ensure for numpy and other numeric types
            param_type = ParameterType.NUMBER
        elif isinstance(value, dict):
            param_type = ParameterType.OBJECT
        else:
            param_type = ParameterType.STRING
        self._logs.params[key] = Parameter(
            name=key,
            value=value,
            description=description,
            parameter_type=param_type,
        )

    def log_metric(
        self,
        key: str,
        value: Any,
        step: int = 0,
        timestamp: int | None = None,
        *,
        description: str | None = None,
    ):
        """Log a metric to the experiment run.

        Args:
            key: Name of the metric.
            value: Value of the metric.
            step: Step number for multi-step metrics (e.g., training epochs).
            timestamp: Unix timestamp in milliseconds when the metric was recorded.

        Keyword Args:
            description: Description of the metric.
        """
        self._logs.metrics[key] = Metric(
            name=key,
            value=value,
            step=step,
            state=ArtifactState.LIVE,
            timestamp=timestamp or str(int(time.time() * 1000)),  # type: ignore[arg-type]
            description=description,
        )

    def log_dataset(
        self,
        name: str | None = None,
        uri: str | None = None,
        source_type: str | None = None,
        source: str | None = None,
        schema: dict | None = None,
        profile: dict | None = None,
        *,
        description: str | None = None,
        file_path: str | None = None,
        s3_auth: S3Params | None = None,
    ):
        """Log a dataset to the experiment run.

        Args:
            name: The name of the dataset.
            uri: The uri of the dataset if already uploaded to S3.
            source_type: Type of the source for the dataset.
            source: Location or connection string for the dataset source.
            schema: JSON schema or description of the dataset structure.
            profile: Statistical profile or summary of the dataset.

        Keyword Args:
            description: Description of the dataset.
            file_path: The path to the dataset file.
            s3_auth: S3Params for uploading the dataset to S3.
        """
        if not uri and not file_path:
            msg = "Either `uri` or `file_path` must be provided."
            raise ValueError(msg)
        try:
            uri = (
                upload_to_s3(
                    s3_auth=s3_auth,  # type: ignore[arg-type]
                    path=file_path,
                )
                if file_path
                else uri
            )
        except Exception as e:
            msg = f"Failed to upload dataset to S3: {e}"
            raise StoreError(msg) from e
        self._logs.datasets[name] = DataSet(  # type: ignore[index]
            name=name,
            uri=uri,
            source_type=source_type,
            source=source,
            schema=schema,  # type: ignore[arg-type]
            profile=profile,  # type: ignore[arg-type]
            description=description,
        )

    def get_log(self, type: LogType, key: str) -> ExperimentRunArtifact:
        """Get a log from the experiment run.

        Args:
            type: Type of the log (params, metrics, or datasets).
            key: Key of the log.
        """
        return self._logs.__getattribute__(type)[key]

    def get_logs(self) -> list[ExperimentRunArtifact]:
        """Return every recorded artifact (params + metrics) in one flat list."""
        params = self._logs.params.values()
        metrics = self._logs.metrics.values()
        datasets = self._logs.datasets.values()

        return list(params) + list(metrics) + list(datasets)



"""Standard client for the model registry."""

from __future__ import annotations

import contextlib
import inspect
import logging
import os
from collections.abc import Callable, Coroutine, Mapping
from dataclasses import asdict
from pathlib import Path
from typing import (
    Any,
    TypeVar,
    get_args,
    overload,
)
from warnings import warn

from model_registry.types.artifacts import ExperimentRunArtifact
from model_registry.types.base import BaseResourceModel

from ._experiments import ActiveExperimentRun, RunContext
from .core import ModelRegistryAPIClient
from .exceptions import StoreError
from .types import (
    Experiment,
    ExperimentRun,
    ListOptions,
    ModelArtifact,
    ModelVersion,
    Pager,
    RegisteredModel,
    SupportedTypes,
)
from .utils import (
    OCIParams,
    S3Params,
    ThreadSafeVariable,
    _connect_to_s3,
    _s3_creds,
    _upload_to_s3,
    generate_name,
    required_args,
    save_to_oci_registry,
)

TModel = TypeVar("TModel", bound=BaseResourceModel)
T = TypeVar("T")

logging.basicConfig(
    format="%(asctime)s.%(msecs)03d - %(name)s:%(levelname)s: %(message)s",
    datefmt="%H:%M:%S",
    level=logging.WARNING,  # the default loglevel
    handlers=[
        # logging.FileHandler(
        #     LOGS
        #     / "log-{}-{}.log".format(
        #         datetime.now(tz=datetime.now().astimezone().tzinfo).strftime(
        #             "%Y-%m-%d-%H-%M-%S"
        #         ),
        #         os.getpid(),
        #     ),
        #     encoding="utf-8",
        #     delay=False,
        # ),
        logging.StreamHandler(),
    ],
)

logger = logging.getLogger("model-registry")

DEFAULT_USER_TOKEN_ENVVAR = "KF_PIPELINES_SA_TOKEN_PATH"  # noqa: S105
DEFAULT_K8S_SA_TOKEN_PATH = "/var/run/secrets/kubernetes.io/serviceaccount/token"  # noqa: S105


class ModelRegistry:
    """Model registry client."""

    def __init__(  # noqa: C901
        self,
        server_address: str,
        port: int = 443,
        *,
        author: str,
        is_secure: bool = True,
        user_token: str | None = None,
        user_token_envvar: str | None = None,
        custom_ca: str | None = None,
        custom_ca_envvar: str | None = None,
        log_level: int = logging.WARNING,
        async_runner: Callable[[Coroutine[Any, Any, T]], T] | None = None,
    ):
        """Constructor.

        Args:
            server_address: Server address.
            port: Server port. Defaults to 443.

        Keyword Args:
            author: Name of the author.
            is_secure: Whether to use a secure connection. Defaults to True.
            user_token: The PEM-encoded user token as a string.
            user_token_envvar: Environment variable to read the user token from if it's not passed as an arg. Defaults to KF_PIPELINES_SA_TOKEN_PATH.
            custom_ca: Path to the PEM-encoded root certificates as a string.
            custom_ca_envvar: Environment variable to read the custom CA from if it's not passed as an arg.
            log_level: Log level. Defaults to logging.WARNING.
            async_runner: A modular async scheduler Callable (either a method or function) that takes in a coroutine for scheduling.
        """
        logger.setLevel(log_level)

        # TODO: get remaining args from env
        self._author = author
        # Set the user's defined async runner
        if async_runner:
            if not (inspect.ismethod(async_runner) or inspect.isfunction(async_runner)):
                msg = "`async_runner` must be a bound method or a function that takes in a coroutine to run."
                raise ValueError(msg)
            self._user_async_runner = async_runner
        else:
            import nest_asyncio

            logger.debug("Setting up reentrant async event loop")
            nest_asyncio.apply()

        if not user_token:
            user_token = self._get_user_token(user_token_envvar)
            if not user_token:
                warn("User access token is missing", stacklevel=2)

        self.hint_server_address_port(server_address, port)
        if is_secure:
            if not custom_ca and custom_ca_envvar and (cert := os.getenv(custom_ca_envvar)):
                logger.info(
                    "Using custom CA envvar %s",
                    custom_ca_envvar,
                )
                custom_ca = cert
                # client might have a default CA setup

            if not user_token:
                msg = "user token must be provided for secure connection"
                raise StoreError(msg)

            self._api = ModelRegistryAPIClient.secure_connection(
                server_address, port, user_token=user_token, custom_ca=custom_ca
            )
        else:
            self._api = ModelRegistryAPIClient.insecure_connection(server_address, port, user_token)
        self._active_experiment_context = ThreadSafeVariable(value=RunContext())
        self.get_registered_models().page_size(1)._next_page()

    @staticmethod
    def _get_user_token(user_token_envvar: str | None = None) -> str | None:
        sa_token_path: str
        user_provided: bool = True
        if user_token_envvar:
            try:
                sa_token_path = os.environ[user_token_envvar]
            except KeyError:
                msg = f"user_token_envvar is {user_token_envvar!r} but no such env var is set"
                raise ValueError(msg) from None
            logger.info(
                "Reading user token from path: user_token_envvar %r specifies path %r",
                user_token_envvar,
                sa_token_path,
            )
        elif DEFAULT_USER_TOKEN_ENVVAR in os.environ:
            sa_token_path = os.environ[DEFAULT_USER_TOKEN_ENVVAR]
            logger.info(
                "Reading user token from path: The default user token env var value %r specifies path %r",
                DEFAULT_USER_TOKEN_ENVVAR,
                sa_token_path,
            )
        else:
            sa_token_path = DEFAULT_K8S_SA_TOKEN_PATH
            user_provided = False
            logger.info(
                "Reading user token from path: No user_token_envvar. Attempting to read from default K8s service account path %r",
                DEFAULT_K8S_SA_TOKEN_PATH,
            )
        try:
            return Path(sa_token_path).read_text()
        except OSError as exc:
            msg = f"Unable read user token from {sa_token_path!r}"
            if user_provided:
                raise StoreError(msg) from exc
            logger.info(msg)
        return None

    @staticmethod
    def hint_server_address_port(server_address: str, port: int) -> None:
        """Hint based on server_address protocol if the port may not be the correct one."""
        if server_address.startswith("https://") and not str(port).endswith("443"):
            logger.warning(
                "Server address protocol is https://, but port is not 443 or ending with 443. You may want to verify the configuration is correct."
            )
        if server_address.startswith("http://") and not str(port).endswith("80"):
            logger.warning(
                "Server address protocol is http://, but port is not 80 or ending with 80. You may want to verify the configuration is correct."
            )

    @overload
    def async_runner(self, coro: Coroutine[Any, Any, TModel]) -> TModel: ...

    @overload
    def async_runner(self, coro: Coroutine[Any, Any, list[TModel]]) -> list[TModel]: ...

    @overload
    def async_runner(self, coro: Coroutine[Any, Any, TModel | None]) -> TModel | None: ...

    def async_runner(self, coro: Coroutine[Any, Any, T]) -> T:
        if hasattr(self, "_user_async_runner"):
            return self._user_async_runner(coro)

        import asyncio

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        return loop.run_until_complete(coro)

    async def _register_model(self, name: str, **kwargs) -> RegisteredModel:
        if rm := await self._api.get_registered_model_by_params(name):
            return rm

        return await self._api.upsert_registered_model(RegisteredModel(name=name, **kwargs))

    async def _register_new_version(self, rm: RegisteredModel, version: str, author: str, /, **kwargs) -> ModelVersion:
        assert rm.id is not None, "Registered model must have an ID"
        if await self._api.get_model_version_by_params(rm.id, version):
            msg = f"Version {version} already exists"
            raise StoreError(msg)

        return await self._api.upsert_model_version(ModelVersion(name=version, author=author, **kwargs), rm.id)

    async def _register_model_artifact(self, mv: ModelVersion, name: str, uri: str, /, **kwargs) -> ModelArtifact:
        assert mv.id is not None, "Model version must have an ID"
        return await self._api.upsert_model_version_artifact(ModelArtifact(name=name, uri=uri, **kwargs), mv.id)

    def upload_artifact_and_register_model(
        self,
        name: str,
        model_files_path: str,
        *,
        # Upload/client Params
        upload_params: OCIParams | S3Params,
        # Artifact/Model Params
        version: str,
        model_format_name: str,
        model_format_version: str,
        storage_path: str | None = None,
        storage_key: str | None = None,
        service_account_name: str | None = None,
        author: str | None = None,
        owner: str | None = None,
        version_description: str | None = None,
        description: str | None = None,
        metadata: Mapping[str, SupportedTypes] | None = None,
    ) -> RegisteredModel:
        """Convenience method to perform 2 operations; uploading an artifact to a storage location, and registers the model in model registry.

        Args:
            name: Name of the model.
            model_files_path: The path where the model files are located. If a directory, uploads the entire directory.

        Keyword Args:
            upload_params: Parameters to configure which storage client to use as well as that client's configuration when uploading the model.
            version: Version of the model. Has to be unique.
            model_format_name: Name of the model format.
            model_format_version: Version of the model format.
            version_description: Description of the model version.
            description: Deprecated. Use version_description instead.
            author: Author of the model. Defaults to the client author.
            owner: Owner of the model. Defaults to the client author.
            storage_key: Storage key.
            storage_path: Storage path.
            service_account_name: Service account name.
            metadata: Additional version metadata. Defaults to values returned by `default_metadata()`.

        Raises:
            ValueError: When the provided `upload_params` is missing or invalid

        Returns:
            Registered model. See: :meth:`~ModelRegistry.register_model`
        """
        # Handle deprecated parameter
        if description is not None:
            warn(
                "The 'description' parameter is deprecated and will be removed in a future version. "
                "Use 'version_description' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            if version_description is None:
                version_description = description

        # Check if model does not already exist in Registry
        ver = None
        with contextlib.suppress(StoreError):
            ver = self.get_model_version(name=name, version=version)

        if ver:
            msg = f"Model `{name}:{version}` already exists in Model Registry."
            raise StoreError(msg)

        if isinstance(upload_params, S3Params):
            destination_uri = self.save_to_s3(**asdict(upload_params), path=model_files_path)
        elif isinstance(upload_params, OCIParams):
            dict_params = asdict(upload_params)
            del dict_params["custom_oci_backend"]
            destination_uri = save_to_oci_registry(
                **dict_params,
                custom_oci_backend=upload_params.custom_oci_backend,
                model_files_path=model_files_path,
            )
        else:
            msg = 'Param "upload_params" is required to perform an upload. Please ensure the value provided is valid'
            raise ValueError(msg)

        return self.register_model(
            name,
            destination_uri,
            model_format_name=model_format_name,
            model_format_version=model_format_version,
            version=version,
            storage_key=storage_key,
            storage_path=storage_path,
            service_account_name=service_account_name,
            author=author,
            owner=owner,
            version_description=version_description,
            metadata=metadata,
        )

    def register_model(
        self,
        name: str,
        uri: str,
        *,
        model_format_name: str,
        model_format_version: str,
        version: str,
        storage_key: str | None = None,
        storage_path: str | None = None,
        service_account_name: str | None = None,
        model_source_kind: str | None = None,
        model_source_class: str | None = None,
        model_source_group: str | None = None,
        model_source_id: str | None = None,
        model_source_name: str | None = None,
        author: str | None = None,
        owner: str | None = None,
        version_description: str | None = None,
        description: str | None = None,
        metadata: Mapping[str, SupportedTypes] | None = None,
    ) -> RegisteredModel:
        """Register a model.

        This registers a model in the model registry. The model is not downloaded, and has to be stored prior to
        registration.

        Most models can be registered using their URI, along with optional connection-specific parameters, `storage_key`
        and `storage_path` or, simply a `service_account_name`.
        URI builder utilities are recommended when referring to specialized storage; for example `utils.s3_uri_from`
        helper when using S3 object storage data connections.

        Args:
            name: Name of the model.
            uri: URI of the model.

        Keyword Args:
            version: Version of the model. Has to be unique.
            model_format_name: Name of the model format.
            model_format_version: Version of the model format.
            version_description: Description of the model version.
            description: Deprecated. Use version_description instead.
            author: Author of the model. Defaults to the client author.
            owner: Owner of the model. Defaults to the client author.
            storage_key: Storage key.
            storage_path: Storage path.
            service_account_name: Service account name.
            model_source_kind: A string identifier describing the source kind.
            model_source_class: A subgroup within the source kind.
            model_source_group: This identifies a source group for models from source class.
            model_source_id: A unique identifier for a source model within kind, class, and group.
            model_source_name: A human-readable name for the source model.
            metadata: Additional version metadata. Defaults to values returned by `default_metadata()`.

        Returns:
            Registered model.
        """
        # Handle deprecated parameter
        if description is not None:
            warn(
                "The 'description' parameter is deprecated and will be removed in a future version. "
                "Use 'version_description' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            if version_description is None:
                version_description = description

        rm = self.async_runner(self._register_model(name, owner=owner or self._author))
        mv = self.async_runner(
            self._register_new_version(
                rm,
                version,
                author or self._author,
                description=version_description,
                custom_properties=metadata or {},
            )
        )
        self.async_runner(
            self._register_model_artifact(
                mv,
                name,
                uri,
                model_format_name=model_format_name,
                model_format_version=model_format_version,
                storage_key=storage_key,
                storage_path=storage_path,
                service_account_name=service_account_name,
                model_source_kind=model_source_kind,
                model_source_class=model_source_class,
                model_source_group=model_source_group,
                model_source_id=model_source_id,
                model_source_name=model_source_name,
            )
        )

        return rm

    def update(self, model: TModel) -> TModel:
        """Update a model."""
        if not model.id:
            msg = "Model must have an ID"
            raise StoreError(msg)
        if not isinstance(model, BaseResourceModel):
            msg = f"Model must be an instance of {BaseResourceModel.__name__} or a subclass"
            raise StoreError(msg)
        if isinstance(model, RegisteredModel):
            return self.async_runner(self._api.upsert_registered_model(model))  # type: ignore[return-value]
        if isinstance(model, ModelVersion):
            return self.async_runner(self._api.upsert_model_version(model, None))  # type: ignore[return-value]
        return self.async_runner(self._api.upsert_model_artifact(model))  # type: ignore[arg-type,return-value]

    def register_hf_model(  # noqa: C901
        self,
        repo: str,
        path: str,
        *,
        version: str,
        model_format_name: str,
        model_format_version: str,
        author: str | None = None,
        owner: str | None = None,
        model_name: str | None = None,
        version_description: str | None = None,
        description: str | None = None,
        git_ref: str = "main",
    ) -> RegisteredModel:
        """Register a Hugging Face model.

        This imports a model from Hugging Face hub and registers it in the model registry.
        Note that the model is not downloaded.

        Args:
            repo: Name of the repository from Hugging Face hub.
            path: URI of the model.

        Keyword Args:
            version: Version of the model. Has to be unique.
            model_format_name: Name of the model format.
            model_format_version: Version of the model format.
            author: Author of the model. Defaults to repo owner.
            owner: Owner of the model. Defaults to the client author.
            model_name: Name of the model. Defaults to the repo name.
            version_description: Description of the model version.
            description: Deprecated. Use version_description instead.
            git_ref: Git reference to use. Defaults to `main`.

        Returns:
            Registered model.
        """
        # Handle deprecated parameter
        if description is not None:
            warn(
                "The 'description' parameter is deprecated and will be removed in a future version. "
                "Use 'version_description' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            if version_description is None:
                version_description = description

        try:
            from huggingface_hub import HfApi, hf_hub_url, utils
        except ImportError as e:
            msg = """package `huggingface-hub` is not installed.
            To import models from Hugging Face Hub, start by installing the `huggingface-hub` package,
            either directly or as an extra (available as `model-registry[hf]`), e.g.:
            ```sh
            !pip install --pre model-registry[hf]
            ```
            or
            ```sh
            !pip install huggingface-hub
            ```
            """
            raise StoreError(msg) from e

        api = HfApi()
        try:
            model_info = api.model_info(repo, revision=git_ref)
        except utils.RepositoryNotFoundError as e:
            msg = f"Repository {repo} does not exist"
            raise StoreError(msg) from e
        except utils.RevisionNotFoundError as e:
            # TODO: as all hf-hub client calls default to using main, should we provide a tip?
            msg = f"Revision {git_ref} does not exist"
            raise StoreError(msg) from e

        if not author:
            # model author can be None if the repo is in a "global" namespace (i.e. no / in repo).
            if model_info.author is None:
                model_author = "unknown"
                warn(
                    "Model author is unknown. This is likely because the model is in a global namespace.",
                    stacklevel=2,
                )
            else:
                model_author = model_info.author
        else:
            model_author = author
        source_uri = hf_hub_url(repo, path, revision=git_ref)
        metadata = {
            "repo": repo,
            "source_uri": source_uri,
            "model_origin": "huggingface_hub",
            "model_author": model_author,
        }
        # card_data is the new field, but let's use the old one for backwards compatibility.
        if card_data := model_info.cardData:
            metadata.update(
                {
                    k: v
                    for k, v in card_data.to_dict().items()
                    # TODO: (#151) preserve tags, possibly other complex metadata
                    if isinstance(v, get_args(SupportedTypes))
                }
            )
        return self.register_model(
            model_name or model_info.id,
            source_uri,
            author=author or model_author,
            owner=owner or self._author,
            version=version,
            model_format_name=model_format_name,
            model_format_version=model_format_version,
            version_description=version_description,
            storage_path=path,
            metadata=metadata,
        )

    def get_registered_model(self, name: str) -> RegisteredModel | None:
        """Get a registered model.

        Args:
            name: Name of the model.

        Returns:
            Registered model.
        """
        return self.async_runner(self._api.get_registered_model_by_params(name))

    def get_model_version(self, name: str, version: str) -> ModelVersion | None:
        """Get a model version.

        Args:
            name: Name of the model.
            version: Version of the model.

        Returns:
            Model version.

        Raises:
            StoreException: If the model does not exist.
        """
        if not (rm := self.get_registered_model(name)):
            msg = f"Model {name} does not exist"
            raise StoreError(msg)
        assert rm.id
        return self.async_runner(self._api.get_model_version_by_params(rm.id, version))

    def get_model_artifact(self, name: str, version: str) -> ModelArtifact | None:
        """Get a model artifact.

        Args:
            name: Name of the model.
            version: Version of the model.

        Returns:
            Model artifact.

        Raises:
            StoreException: If either the model or the version don't exist.
        """
        if not (mv := self.get_model_version(name, version)):
            msg = f"Version {version} does not exist"
            raise StoreError(msg)
        assert mv.id
        return self.async_runner(self._api.get_model_artifact_by_params(name, mv.id))

    def get_registered_models(self) -> Pager[RegisteredModel]:
        """Get a pager for registered models.

        Returns:
            Iterable pager for registered models.
        """

        def rm_list(options: ListOptions) -> list[RegisteredModel]:
            return self.async_runner(self._api.get_registered_models(options))

        return Pager[RegisteredModel](rm_list)

    def get_model_versions(self, name: str) -> Pager[ModelVersion]:
        """Get a pager for model versions.

        Args:
            name: Name of the model.

        Returns:
            Iterable pager for model versions.

        Raises:
            StoreException: If the model does not exist.
        """
        if not (rm := self.get_registered_model(name)):
            msg = f"Model {name} does not exist"
            raise StoreError(msg)

        def rm_versions(options: ListOptions) -> list[ModelVersion]:
            # type checkers can't restrict the type inside a nested function:
            # https://mypy.readthedocs.io/en/stable/common_issues.html#narrowing-and-inner-functions
            assert rm.id
            return self.async_runner(self._api.get_model_versions(rm.id, options))

        return Pager[ModelVersion](rm_versions)

    def save_to_s3(
        self,
        path: str,
        bucket_name: str,
        s3_prefix: str,
        *,
        endpoint_url: str | None = None,
        access_key_id: str | None = None,
        secret_access_key: str | None = None,
        region: str | None = None,
        multipart_threshold: int = 1024 * 1024,
        multipart_chunksize: int = 1024 * 1024,
        max_pool_connections: int = 10,
    ) -> str:
        """Saves a model to an S3 compatible storage.

        Args:
            path: Location to where the model(s) or artifact(s) are located. \
                Can recursively upload nested files in folders.
            bucket_name: The bucket to use for the S3 compatible object storage.
            s3_prefix: The path to prefix under root of bucket.

        Keyword Args:
            endpoint_url: The endpoint URL for the S3 comaptible storage if not using AWS S3.
            access_key_id: The S3 compatible object storage access ID.
            secret_access_key: The S3 compatible object storage secret access key.
            region: The region name for the S3 object storage.
            multipart_threshold: The threshold for multipart uploads.
            multipart_chunksize: The size of chunks for multipart uploads.
            max_pool_connections: The maximum number of connections in the pool.

        Returns:
            The S3 URI to the uploaded files.

        Raises:
            StoreError: If there was an issue uploading to S3.
        """
        # Get mixed credentials
        endpoint_url, access_key_id, secret_access_key, region = _s3_creds(
            endpoint_url, access_key_id, secret_access_key, region
        )

        s3, transfer_config = _connect_to_s3(
            endpoint_url=endpoint_url,
            access_key_id=access_key_id,
            secret_access_key=secret_access_key,
            multipart_threshold=multipart_threshold,
            multipart_chunksize=multipart_chunksize,
            max_pool_connections=max_pool_connections,
        )
        return _upload_to_s3(
            path=path,
            path_prefix=s3_prefix,
            bucket=bucket_name,
            s3=s3,
            endpoint_url=endpoint_url,
            region=region,
            transfer_config=transfer_config,
        )

    def start_experiment_run(
        self,
        experiment_name: str | None = None,
        experiment_id: str | None = None,
        run_name: str | None = None,
        run_id: str | None = None,
        *,
        owner: str | None = None,
        description: str | None = None,
        run_description: str | None = None,
        nested: bool = False,
        nested_tag: str | None = "kubeflow.parent_run_id",
    ) -> ActiveExperimentRun:
        """Start an experiment run.

        Args:
            experiment_name: Name of the experiment.
            experiment_id: ID of the experiment.
            run_name: Name of the run.
            run_id: ID of the run.

        Keyword Args:
            owner: Owner of the experiment.
            description: Description of the experiment.
            run_description: Description of the run.
            nested: Whether the run is nested.
            nested_tag: Tag to use for nested runs.

        Returns:
            Experiment run.
        """
        active_ctx = self._get_active_context()
        self._validate_nested_run(active_ctx, nested)

        # Resolve experiment details
        exp_name, exp_id = self._resolve_experiment_info(experiment_name, experiment_id, active_ctx, nested)

        # Get or create experiment
        experiment = self._get_or_create_experiment(exp_name, exp_id, owner, description)

        # Get or create run
        parent_props = self._get_parent_properties(active_ctx, nested_tag) if nested else {}  # type: ignore[arg-type]
        exp_run = self._get_or_create_run(experiment, run_name, run_id, run_description, parent_props, nested)

        # Update context if not nested
        if not active_ctx.active:
            self._set_active_context(experiment.id, exp_name, exp_run.id)  # type: ignore[arg-type]

        return ActiveExperimentRun(
            thread_safe_ctx=self._active_experiment_context,
            experiment_run=exp_run,
            api=self._api,
            async_runner=self.async_runner,
        )

    def _get_active_context(self) -> RunContext:
        """Get the current active experiment context."""
        return self._active_experiment_context.get()

    def _validate_nested_run(self, active_ctx: RunContext, nested: bool) -> None:
        """Validate nested run configuration."""
        if active_ctx.active and not nested:
            msg = "Experiment run is already active. Please set nested=True to start a nested run."
            raise ValueError(msg)

    def _resolve_experiment_info(
        self,
        exp_name: str | None,
        exp_id: str | None,
        active_ctx: RunContext,
        nested: bool,
    ) -> tuple[str | None, str | None]:
        """Resolve experiment name and ID from inputs or context."""
        # Use provided values or inherit from active context
        exp_name = exp_name or active_ctx.name
        exp_id = exp_id or active_ctx.id

        # Generate name if nothing provided and not nested
        if not any([exp_name, exp_id, nested]):
            exp_name = generate_name("experiment")

        return exp_name, exp_id

    def _get_or_create_experiment(
        self,
        exp_name: str | None,
        exp_id: str | None,
        owner: str | None,
        description: str | None,
    ) -> Experiment:
        """Get existing experiment or create new one."""
        # Try to get existing experiment
        if exp_name:
            exp = self.async_runner(self._api.get_experiment_by_name(exp_name))
        elif exp_id:
            exp = self.async_runner(self._api.get_experiment_by_id(exp_id))
        else:
            msg = "Either experiment_name or experiment_id must be provided"
            raise ValueError(msg)

        # Create if doesn't exist
        if not exp:
            exp = self.async_runner(
                self._api.upsert_experiment(
                    Experiment(
                        name=exp_name,  # type: ignore[arg-type]
                        owner=owner,
                        description=description,
                    )
                )
            )
            print(f"Experiment {exp_name} created with ID: {exp.id}")

        return exp

    def _get_parent_properties(self, active_ctx: RunContext, nested_tag: str) -> dict:
        """Get parent run properties for nested runs."""
        return {nested_tag: active_ctx.run_id} if active_ctx.active else {}

    def _get_or_create_run(
        self,
        experiment: Experiment,
        run_name: str | None,
        run_id: str | None,
        run_description: str | None,
        parent_props: dict,
        nested: bool,
    ) -> ExperimentRun:
        """Get existing run or create new one."""
        exp_run_args = {
            "experiment_name": experiment.name,
            "experiment_id": experiment.id,
        }

        # Try to get existing run
        if run_name:
            exp_run = self.async_runner(
                self._api.get_experiment_run_by_experiment_and_run_name(
                    run_name=run_name,
                    **exp_run_args,  # type: ignore[arg-type]
                )
            )
        elif run_id:
            exp_run = self.async_runner(
                self._api.get_experiment_run_by_experiment_and_run_id(
                    run_id=run_id,
                    **exp_run_args,
                )
            )
        else:
            # Create new run
            exp_run = self.async_runner(
                self._api.upsert_experiment_run(
                    ExperimentRun(
                        experiment_id=experiment.id,  # type: ignore[arg-type]
                        name=generate_name("run"),
                        description=run_description,
                        custom_properties=parent_props,
                    )
                )
            )
            prefix = "Nested " if nested else ""
            print(f"{prefix}Experiment Run {exp_run.name} created with ID: {exp_run.id}")

        return exp_run

    def _set_active_context(self, exp_id: str, exp_name: str, run_id: str) -> None:
        """Set the active experiment context."""
        new_ctx = RunContext(id=exp_id, name=exp_name, run_id=run_id, active=True)
        self._active_experiment_context.set(new_ctx)

    def create_experiment(self, name: str) -> Experiment:
        """Create an experiment.

        Args:
            name: Name of the experiment.
        """
        return self.async_runner(self._api.upsert_experiment(Experiment(name=name)))

    def get_experiment_run(self, run_id: str) -> ExperimentRun:
        """Get an experiment run.

        Args:
            run_id: ID of the experiment run.
        """
        return self.async_runner(self._api.get_experiment_run_by_id(run_id))

    def get_experiments(self) -> Pager[Experiment]:
        """Get a pager for experiments.

        Returns:
            Iterable pager for experiments.
        """

        def exp_list(options: ListOptions) -> list[Experiment]:
            return self.async_runner(self._api.get_experiments(options))

        return Pager[Experiment](exp_list)

    @overload
    def get_experiment_runs(self, experiment_id: str) -> Pager[ExperimentRun]: ...

    @overload
    def get_experiment_runs(self, experiment_name: str) -> Pager[ExperimentRun]: ...

    @required_args(("experiment_id",), ("experiment_name",))  # type: ignore[misc]
    def get_experiment_runs(
        self, experiment_id: str | None = None, experiment_name: str | None = None
    ) -> Pager[ExperimentRun]:
        """Get a pager for experiment runs.

        Returns:
            Iterable pager for experiment runs.
        """

        def exp_run_list(options: ListOptions) -> list[ExperimentRun]:
            if experiment_id:
                return self.async_runner(self._api.get_experiment_runs_by_experiment_id(experiment_id, options))
            return self.async_runner(self._api.get_experiment_runs_by_experiment_name(experiment_name, options))  # type: ignore[arg-type,type-var]

        return Pager[ExperimentRun](exp_run_list)

    @overload
    def get_experiment_run_logs(
        self,
        run_id: str,
    ) -> Pager[ExperimentRunArtifact]: ...

    @overload
    def get_experiment_run_logs(
        self,
        run_name: str,
        experiment_name: str,
    ) -> Pager[ExperimentRunArtifact]: ...

    @overload
    def get_experiment_run_logs(
        self,
        run_name: str,
        experiment_id: str,
    ) -> Pager[ExperimentRunArtifact]: ...

    @required_args(  # type: ignore[misc]
        ("run_id",),
        (
            "run_name",
            "experiment_name",
        ),
        (
            "run_name",
            "experiment_id",
        ),
    )
    def get_experiment_run_logs(
        self,
        run_id: str | None = None,
        run_name: str | None = None,
        experiment_id: str | None = None,
        experiment_name: str | None = None,
    ) -> Pager[ExperimentRunArtifact]:
        """Get a pager for experiment run logs.

        Args:
            run_id: ID of the experiment run.
            run_name: Name of the experiment run.
            experiment_id: ID of the experiment.
            experiment_name: Name of the experiment.

        Returns:
            Iterable pager for experiment run logs.
        """

        def exp_run_logs(options: ListOptions) -> list[ExperimentRunArtifact]:
            if run_id:
                return self.async_runner(
                    self._api.get_artifacts_by_experiment_run_params(run_id=run_id, options=options)
                )
            if run_name and experiment_name:
                return self.async_runner(
                    self._api.get_artifacts_by_experiment_run_params(
                        run_name=run_name,
                        experiment_name=experiment_name,
                        options=options,
                    )
                )
            if run_name and experiment_id:
                return self.async_runner(
                    self._api.get_artifacts_by_experiment_run_params(
                        run_name=run_name, experiment_id=experiment_id, options=options
                    )
                )
            return None  # type: ignore[return-value]

        return Pager[ExperimentRunArtifact](exp_run_logs)

    # TODO: consider porting get_artifacts method here
    # https://github.com/kubeflow/model-registry/pull/1536



"""Experiment types for model registry.

Experiment types are used to map between proto types based on experiments and Python objects.


"""

from __future__ import annotations  # noqa: I001

from typing import Any
from typing_extensions import override

from mr_openapi import (
    Experiment as ExperimentBaseModel,
)
from mr_openapi import (
    ExperimentCreate,
    ExperimentRunCreate,
    ExperimentRunState,
    ExperimentRunUpdate,
    ExperimentState,
    ExperimentUpdate,
)
from mr_openapi import (
    ExperimentRun as ExperimentRunBaseModel,
)

from .base import BaseResourceModel


class Experiment(BaseResourceModel):
    """Represents an experiment model.

    Attributes:
        name: Name of the experiment.
        owner: Owner of the experiment.
        description: Description of the experiment.
        external_id: External ID of the experiment.
        state: State of the experiment.
        custom_properties: Custom properties (metadata)of the experiment.
    """

    name: str
    owner: str | None = None
    description: str | None = None
    external_id: str | None = None
    state: ExperimentState | None = None
    custom_properties: dict[str, Any] | None = None

    @override
    def create(self, **kwargs) -> ExperimentCreate:
        return ExperimentCreate(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("id", "custom_properties")),
            **kwargs,
        )

    @override
    def update(self, **kwargs) -> ExperimentUpdate:
        return ExperimentUpdate(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("id", "name", "custom_properties")),
            **kwargs,
        )

    @classmethod
    @override
    def from_basemodel(cls, source: ExperimentBaseModel) -> Experiment:
        assert source.name
        return cls(
            id=source.id,
            name=source.name,
            owner=source.owner,
            description=source.description,
            external_id=source.external_id,
            create_time_since_epoch=source.create_time_since_epoch,
            last_update_time_since_epoch=source.last_update_time_since_epoch,
            state=source.state,
            custom_properties=cls._unmap_custom_properties(source.custom_properties)
            if source.custom_properties
            else None,
        )


class ExperimentRun(BaseResourceModel):
    """Represents an experiment run model.

    Attributes:
        name: Name of the experiment run.
        owner: Owner of the experiment run.
        description: Description of the experiment run.
        external_id: External ID of the experiment run.
        state: State of the experiment.
        custom_properties: Custom properties (metadata)of the experiment.
    """

    name: str
    experiment_id: str
    owner: str | None = None
    description: str | None = None
    external_id: str | None = None
    state: ExperimentRunState | None = None
    custom_properties: dict[str, Any] | None = None

    @override
    def create(self, **kwargs) -> ExperimentRunCreate:
        return ExperimentRunCreate(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("id", "custom_properties")),
            **kwargs,
        )

    @override
    def update(self, **kwargs) -> ExperimentRunUpdate:
        return ExperimentRunUpdate(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("id", "name", "custom_properties")),
            **kwargs,
        )

    @classmethod
    @override
    def from_basemodel(cls, source: ExperimentRunBaseModel) -> ExperimentRun:
        assert source.name
        return cls(
            id=source.id,
            name=source.name,
            experiment_id=source.experiment_id,
            owner=source.owner,
            description=source.description,
            external_id=source.external_id,
            create_time_since_epoch=source.create_time_since_epoch,
            last_update_time_since_epoch=source.last_update_time_since_epoch,
            state=source.state,
            custom_properties=cls._unmap_custom_properties(source.custom_properties)
            if source.custom_properties
            else None,
        )



# coding: utf-8

# flake8: noqa

"""
Model Registry REST API

REST API for Model Registry to create and manage ML model metadata

The version of the OpenAPI document: v1alpha3
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

__version__ = "1.0.0"

# Define package exports
__all__ = [
    "ModelRegistryServiceApi",
    "ApiResponse",
    "ApiClient",
    "Configuration",
    "OpenApiException",
    "ApiTypeError",
    "ApiValueError",
    "ApiKeyError",
    "ApiAttributeError",
    "ApiException",
    "Artifact",
    "ArtifactCreate",
    "ArtifactList",
    "ArtifactState",
    "ArtifactTypeQueryParam",
    "ArtifactUpdate",
    "BaseArtifact",
    "BaseModel",
    "BaseResource",
    "BaseResourceCreate",
    "BaseResourceDates",
    "BaseResourceList",
    "BaseResourceUpdate",
    "DataSet",
    "DataSetCreate",
    "DataSetUpdate",
    "DocArtifact",
    "DocArtifactCreate",
    "DocArtifactUpdate",
    "Error",
    "ExecutionState",
    "Experiment",
    "ExperimentCreate",
    "ExperimentList",
    "ExperimentRun",
    "ExperimentRunCreate",
    "ExperimentRunList",
    "ExperimentRunState",
    "ExperimentRunStatus",
    "ExperimentRunUpdate",
    "ExperimentState",
    "ExperimentUpdate",
    "InferenceService",
    "InferenceServiceCreate",
    "InferenceServiceList",
    "InferenceServiceState",
    "InferenceServiceUpdate",
    "MetadataBoolValue",
    "MetadataDoubleValue",
    "MetadataIntValue",
    "MetadataProtoValue",
    "MetadataStringValue",
    "MetadataStructValue",
    "MetadataValue",
    "Metric",
    "MetricCreate",
    "MetricList",
    "MetricUpdate",
    "ModelArtifact",
    "ModelArtifactCreate",
    "ModelArtifactList",
    "ModelArtifactUpdate",
    "ModelVersion",
    "ModelVersionCreate",
    "ModelVersionList",
    "ModelVersionState",
    "ModelVersionUpdate",
    "OrderByField",
    "Parameter",
    "ParameterCreate",
    "ParameterType",
    "ParameterUpdate",
    "RegisteredModel",
    "RegisteredModelCreate",
    "RegisteredModelList",
    "RegisteredModelState",
    "RegisteredModelUpdate",
    "ServeModel",
    "ServeModelCreate",
    "ServeModelList",
    "ServeModelUpdate",
    "ServingEnvironment",
    "ServingEnvironmentCreate",
    "ServingEnvironmentList",
    "ServingEnvironmentUpdate",
    "SortOrder",
]

# import apis into sdk package
from mr_openapi.api.model_registry_service_api import ModelRegistryServiceApi as ModelRegistryServiceApi

# import ApiClient
from mr_openapi.api_response import ApiResponse as ApiResponse
from mr_openapi.api_client import ApiClient as ApiClient
from mr_openapi.configuration import Configuration as Configuration
from mr_openapi.exceptions import OpenApiException as OpenApiException
from mr_openapi.exceptions import ApiTypeError as ApiTypeError
from mr_openapi.exceptions import ApiValueError as ApiValueError
from mr_openapi.exceptions import ApiKeyError as ApiKeyError
from mr_openapi.exceptions import ApiAttributeError as ApiAttributeError
from mr_openapi.exceptions import ApiException as ApiException

# import models into sdk package
from mr_openapi.models.artifact import Artifact as Artifact
from mr_openapi.models.artifact_create import ArtifactCreate as ArtifactCreate
from mr_openapi.models.artifact_list import ArtifactList as ArtifactList
from mr_openapi.models.artifact_state import ArtifactState as ArtifactState
from mr_openapi.models.artifact_type_query_param import ArtifactTypeQueryParam as ArtifactTypeQueryParam
from mr_openapi.models.artifact_update import ArtifactUpdate as ArtifactUpdate
from mr_openapi.models.base_artifact import BaseArtifact as BaseArtifact
from mr_openapi.models.base_model import BaseModel as BaseModel
from mr_openapi.models.base_resource import BaseResource as BaseResource
from mr_openapi.models.base_resource_create import BaseResourceCreate as BaseResourceCreate
from mr_openapi.models.base_resource_dates import BaseResourceDates as BaseResourceDates
from mr_openapi.models.base_resource_list import BaseResourceList as BaseResourceList
from mr_openapi.models.base_resource_update import BaseResourceUpdate as BaseResourceUpdate
from mr_openapi.models.data_set import DataSet as DataSet
from mr_openapi.models.data_set_create import DataSetCreate as DataSetCreate
from mr_openapi.models.data_set_update import DataSetUpdate as DataSetUpdate
from mr_openapi.models.doc_artifact import DocArtifact as DocArtifact
from mr_openapi.models.doc_artifact_create import DocArtifactCreate as DocArtifactCreate
from mr_openapi.models.doc_artifact_update import DocArtifactUpdate as DocArtifactUpdate
from mr_openapi.models.error import Error as Error
from mr_openapi.models.execution_state import ExecutionState as ExecutionState
from mr_openapi.models.experiment import Experiment as Experiment
from mr_openapi.models.experiment_create import ExperimentCreate as ExperimentCreate
from mr_openapi.models.experiment_list import ExperimentList as ExperimentList
from mr_openapi.models.experiment_run import ExperimentRun as ExperimentRun
from mr_openapi.models.experiment_run_create import ExperimentRunCreate as ExperimentRunCreate
from mr_openapi.models.experiment_run_list import ExperimentRunList as ExperimentRunList
from mr_openapi.models.experiment_run_state import ExperimentRunState as ExperimentRunState
from mr_openapi.models.experiment_run_status import ExperimentRunStatus as ExperimentRunStatus
from mr_openapi.models.experiment_run_update import ExperimentRunUpdate as ExperimentRunUpdate
from mr_openapi.models.experiment_state import ExperimentState as ExperimentState
from mr_openapi.models.experiment_update import ExperimentUpdate as ExperimentUpdate
from mr_openapi.models.inference_service import InferenceService as InferenceService
from mr_openapi.models.inference_service_create import InferenceServiceCreate as InferenceServiceCreate
from mr_openapi.models.inference_service_list import InferenceServiceList as InferenceServiceList
from mr_openapi.models.inference_service_state import InferenceServiceState as InferenceServiceState
from mr_openapi.models.inference_service_update import InferenceServiceUpdate as InferenceServiceUpdate
from mr_openapi.models.metadata_bool_value import MetadataBoolValue as MetadataBoolValue
from mr_openapi.models.metadata_double_value import MetadataDoubleValue as MetadataDoubleValue
from mr_openapi.models.metadata_int_value import MetadataIntValue as MetadataIntValue
from mr_openapi.models.metadata_proto_value import MetadataProtoValue as MetadataProtoValue
from mr_openapi.models.metadata_string_value import MetadataStringValue as MetadataStringValue
from mr_openapi.models.metadata_struct_value import MetadataStructValue as MetadataStructValue
from mr_openapi.models.metadata_value import MetadataValue as MetadataValue
from mr_openapi.models.metric import Metric as Metric
from mr_openapi.models.metric_create import MetricCreate as MetricCreate
from mr_openapi.models.metric_list import MetricList as MetricList
from mr_openapi.models.metric_update import MetricUpdate as MetricUpdate
from mr_openapi.models.model_artifact import ModelArtifact as ModelArtifact
from mr_openapi.models.model_artifact_create import ModelArtifactCreate as ModelArtifactCreate
from mr_openapi.models.model_artifact_list import ModelArtifactList as ModelArtifactList
from mr_openapi.models.model_artifact_update import ModelArtifactUpdate as ModelArtifactUpdate
from mr_openapi.models.model_version import ModelVersion as ModelVersion
from mr_openapi.models.model_version_create import ModelVersionCreate as ModelVersionCreate
from mr_openapi.models.model_version_list import ModelVersionList as ModelVersionList
from mr_openapi.models.model_version_state import ModelVersionState as ModelVersionState
from mr_openapi.models.model_version_update import ModelVersionUpdate as ModelVersionUpdate
from mr_openapi.models.order_by_field import OrderByField as OrderByField
from mr_openapi.models.parameter import Parameter as Parameter
from mr_openapi.models.parameter_create import ParameterCreate as ParameterCreate
from mr_openapi.models.parameter_type import ParameterType as ParameterType
from mr_openapi.models.parameter_update import ParameterUpdate as ParameterUpdate
from mr_openapi.models.registered_model import RegisteredModel as RegisteredModel
from mr_openapi.models.registered_model_create import RegisteredModelCreate as RegisteredModelCreate
from mr_openapi.models.registered_model_list import RegisteredModelList as RegisteredModelList
from mr_openapi.models.registered_model_state import RegisteredModelState as RegisteredModelState
from mr_openapi.models.registered_model_update import RegisteredModelUpdate as RegisteredModelUpdate
from mr_openapi.models.serve_model import ServeModel as ServeModel
from mr_openapi.models.serve_model_create import ServeModelCreate as ServeModelCreate
from mr_openapi.models.serve_model_list import ServeModelList as ServeModelList
from mr_openapi.models.serve_model_update import ServeModelUpdate as ServeModelUpdate
from mr_openapi.models.serving_environment import ServingEnvironment as ServingEnvironment
from mr_openapi.models.serving_environment_create import ServingEnvironmentCreate as ServingEnvironmentCreate
from mr_openapi.models.serving_environment_list import ServingEnvironmentList as ServingEnvironmentList
from mr_openapi.models.serving_environment_update import ServingEnvironmentUpdate as ServingEnvironmentUpdate
from mr_openapi.models.sort_order import SortOrder as SortOrder



"""Cosign signing and verification for container images."""

import os
import shutil
import subprocess
import sys
from functools import partial
from pathlib import Path

from model_registry.signing.exceptions import InitializationError, SigningError, VerificationError


class CommandRunner:
    """Command execution wrapper for running CLI tools."""

    def __init__(self, **kwargs):
        """Initialize CommandRunner with default subprocess.run arguments.

        Args:
            **kwargs: Additional default keyword arguments to pass to subprocess.run
        """
        self._run = partial(subprocess.run, check=True, capture_output=True, text=True, **kwargs)

    def run(self, cmd: list[str]) -> subprocess.CompletedProcess:
        """Run a CLI command and handle output.

        Args:
            cmd: Command and arguments as a list

        Returns:
            CompletedProcess instance

        Raises:
            subprocess.CalledProcessError: If the command fails
        """
        result = self._run(cmd)
        if result.stderr:
            print(result.stderr, file=sys.stderr)
        return result


class ImageSigner:
    """Tool for signing and verifying container images."""

    def __init__(
        self,
        tuf_url: str | None = None,
        root: str | None = None,
        root_checksum: str | None = None,
        identity_token_path: str | os.PathLike[str] | None = None,
        fulcio_url: str | None = None,
        rekor_url: str | None = None,
        certificate_identity: str | None = None,
        oidc_issuer: str | None = None,
    ):
        """Initialize ImageSigner tool.

        Args:
            tuf_url: Default TUF mirror URL
            root: Default root JSON URL
            root_checksum: Default root JSON checksum
            identity_token_path: Default path to OIDC identity token file
            fulcio_url: Default Fulcio server URL
            rekor_url: Default Rekor server URL
            certificate_identity: Default certificate identity
            oidc_issuer: Default OIDC issuer URL

        Raises:
            FileNotFoundError: If identity_token_path is provided but doesn't exist
        """
        self.runner = CommandRunner()
        self.tuf_url = tuf_url
        self.root = root
        self.root_checksum = root_checksum

        # Validate identity token path if provided
        if identity_token_path is not None and not os.path.exists(identity_token_path):
            msg = f"Identity token file not found: {identity_token_path}"
            raise FileNotFoundError(msg)
        self.identity_token_path = identity_token_path

        self.fulcio_url = fulcio_url
        self.rekor_url = rekor_url
        self.certificate_identity = certificate_identity
        self.oidc_issuer = oidc_issuer

    def _get_sigstore_dir(self) -> Path:
        """Get the sigstore directory path.

        Returns:
            Path to the sigstore directory
        """
        return Path.home() / ".sigstore"

    def initialize(  # noqa: C901
        self,
        tuf_url: str | None = None,
        root: str | None = None,
        root_checksum: str | None = None,
        force: bool = False,
    ):
        """Initialize sigstore configuration.

        Args:
            tuf_url: TUF mirror URL (optional, overrides instance default)
            root: Root JSON URL (optional, overrides instance default)
            root_checksum: Root JSON checksum (optional, overrides instance default)
            force: If True, remove existing sigstore directory; if False, raise error if exists

        Raises:
            FileExistsError: If sigstore directory exists and force=False
        """
        # Check for existing sigstore directory
        sigstore_dir = self._get_sigstore_dir()
        if sigstore_dir.exists():
            if not force:
                msg = f"Sigstore directory already exists: {sigstore_dir}. Use force=True to overwrite."
                raise FileExistsError(msg)
            shutil.rmtree(sigstore_dir)

        # Use method args or fall back to instance attributes
        if tuf_url is None:
            tuf_url = self.tuf_url
        if root is None:
            root = self.root
        if root_checksum is None:
            root_checksum = self.root_checksum

        # Initialize cosign with TUF mirror
        cmd = ["cosign", "initialize"]

        if tuf_url is not None:
            cmd.extend(["--mirror", tuf_url])

        if root is not None:
            cmd.extend(["--root", root])

        if root_checksum is not None:
            cmd.extend(["--root-checksum", root_checksum])

        try:
            self.runner.run(cmd)
        except subprocess.CalledProcessError as e:
            msg = f"Failed to initialize sigstore: {e}"
            raise InitializationError(msg) from e

    def sign(  # noqa: C901
        self,
        image: str,
        identity_token_path: str | os.PathLike[str] | None = None,
        fulcio_url: str | None = None,
        rekor_url: str | None = None,
    ):
        """Sign a container image and upload the signature.

        Args:
            image: Full image reference including digest (e.g., quay.io/user/image@sha256:...)
            identity_token_path: Path to OIDC identity token file (optional, overrides instance default)
            fulcio_url: Fulcio server URL (optional, overrides instance default)
            rekor_url: Rekor server URL (optional, overrides instance default)

        Raises:
            FileNotFoundError: If identity_token_path is provided but doesn't exist
        """
        # Use method args or fall back to instance attributes
        if identity_token_path is None:
            identity_token_path = self.identity_token_path
        if fulcio_url is None:
            fulcio_url = self.fulcio_url
        if rekor_url is None:
            rekor_url = self.rekor_url

        # Validate identity token path if provided
        if identity_token_path is not None and not os.path.exists(identity_token_path):
            msg = f"Identity token file not found: {identity_token_path}"
            raise FileNotFoundError(msg)

        cmd = ["cosign", "sign", "-y"]

        if identity_token_path is not None:
            cmd.extend(["--identity-token", os.fspath(identity_token_path)])

        if fulcio_url is not None:
            cmd.extend(["--fulcio-url", fulcio_url])

        if rekor_url is not None:
            cmd.extend(["--rekor-url", rekor_url])

        cmd.append(image)

        try:
            self.runner.run(cmd)
        except subprocess.CalledProcessError as e:
            msg = f"Failed to sign image {image}: {e}"
            raise SigningError(msg) from e

    def verify(self, image: str, certificate_identity: str | None = None, oidc_issuer: str | None = None):
        """Verify a signed container image.

        Args:
            image: Full image reference including digest (e.g., quay.io/user/image@sha256:...)
            certificate_identity: Expected certificate identity (optional, overrides instance default)
            oidc_issuer: OIDC issuer URL (optional, overrides instance default)
        """
        # Use method args or fall back to instance attributes
        if certificate_identity is None:
            certificate_identity = self.certificate_identity
        if oidc_issuer is None:
            oidc_issuer = self.oidc_issuer

        cmd = ["cosign", "verify"]

        if certificate_identity is not None:
            cmd.extend(["--certificate-identity", certificate_identity])

        if oidc_issuer is not None:
            cmd.extend(["--certificate-oidc-issuer", oidc_issuer])

        cmd.append(image)

        try:
            self.runner.run(cmd)
        except subprocess.CalledProcessError as e:
            msg = f"Failed to verify image {image}: {e}"
            raise VerificationError(msg) from e



"""Sigstore signer implementation for model signing."""
# mypy: ignore-errors

import pathlib

from model_signing._signing.sign_sigstore import _DEFAULT_CLIENT_ID, _DEFAULT_CLIENT_SECRET
from model_signing._signing.sign_sigstore import Signer as BaseSigner
from sigstore import models as sigstore_models
from sigstore import oidc as sigstore_oidc
from sigstore import sign as sigstore_signer


class Signer(BaseSigner):
    """Signing using Sigstore."""

    def __init__(
        self,
        *,
        oidc_issuer: str | None = None,
        use_ambient_credentials: bool = True,
        use_staging: bool = False,
        identity_token: str | None = None,
        force_oob: bool = False,
        client_id: str | None = None,
        client_secret: str | None = None,
        trust_config: pathlib.Path | None = None,
    ):
        """Initializes Sigstore signers.

        Needs to set-up a signing context to use the public goods instance and
        machinery for getting an identity token to use in signing.

        Args:
            oidc_issuer: An optional OpenID Connect issuer to use instead of the
              default production one. Only relevant if `use_staging = False`.
              Default is empty, relying on the Sigstore configuration.
            use_ambient_credentials: Use ambient credentials (also known as
              Workload Identity). Default is True. If ambient credentials cannot
              be used (not available, or option disabled), a flow to get signer
              identity via OIDC will start.
            use_staging: Use staging configurations, instead of production. This
              is supposed to be set to True only when testing. Default is False.
            force_oob: If True, forces an out-of-band (OOB) OAuth flow. If set,
              the OAuth authentication will not attempt to open the default web
              browser. Instead, it will display a URL and code for manual
              authentication. Default is False, which means the browser will be
              opened automatically if possible.
            identity_token: An explicit identity token to use when signing,
              taking precedence over any ambient credential or OAuth workflow.
            client_id: An optional client ID to use when performing OIDC-based
              authentication. This is typically used to identify the
              application making the request to the OIDC provider. If not
              provided, the default client ID configured by Sigstore will be
              used.
            client_secret: An optional client secret to use along with the
              client ID when authenticating with the OIDC provider. This is
              required for confidential clients that need to prove their
              identity to the OIDC provider. If not provided, it is assumed
              that the client is public or the provider does not require a
              secret.
            trust_config: A path to a custom trust configuration. When
              provided, the signature verification process will rely on the
              supplied PKI and trust configurations, instead of the default
              Sigstore setup. If not specified, the default Sigstore
              configuration is used.
        """
        # Initializes the signing and issuer contexts based on provided
        # configuration.
        if use_staging:
            trust_config = sigstore_models.ClientTrustConfig.staging()
        elif trust_config:
            trust_config = sigstore_models.ClientTrustConfig.from_json(trust_config.read_text())
        else:
            trust_config = sigstore_models.ClientTrustConfig.production()

        if not oidc_issuer:
            oidc_issuer = trust_config.signing_config.get_oidc_url()

        self._oidc_issuer = oidc_issuer
        self._signing_context = sigstore_signer.SigningContext.from_trust_config(trust_config)
        self._use_ambient_credentials = use_ambient_credentials
        self._identity_token = identity_token
        self._force_oob = force_oob
        self._client_id = client_id or _DEFAULT_CLIENT_ID
        self._client_secret = client_secret or _DEFAULT_CLIENT_SECRET

    def _get_identity_token(self) -> sigstore_oidc.IdentityToken:
        """Obtains an identity token to use in signing.

        The precedence matches that of sigstore-python:
        1) Explicitly supplied identity token
        2) Ambient credential detected in the environment, if enabled
        3) Interactive OAuth flow
        """
        if self._identity_token:
            return sigstore_oidc.IdentityToken(self._identity_token, self._client_id)
        if self._use_ambient_credentials:
            token = sigstore_oidc.detect_credential(self._client_id)
            if token:
                return sigstore_oidc.IdentityToken(token, self._client_id)

        issuer = sigstore_oidc.Issuer(self._oidc_issuer)
        return issuer.identity_token(
            force_oob=self._force_oob,
            client_id=self._client_id,
            client_secret=self._client_secret,
        )



"""Artifact types for model registry.

Artifacts represent pieces of data.
This could be datasets, models, metrics, or any other piece of data produced or consumed by an
execution, such as an experiment run.

Those types are used to map between proto types based on artifacts and Python objects.

Todo:
    * Move part of the description to API Reference docs (#120).
"""

from __future__ import annotations  # noqa: I001

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, TypeVar

from typing_extensions import override


import json

from mr_openapi import (
    Artifact as ArtifactBaseModel,
)
from mr_openapi import (
    DataSet as DataSetBaseModel,
)
from mr_openapi import (
    Metric as MetricBaseModel,
)
from mr_openapi import (
    Parameter as ParameterBaseModel,
)

from mr_openapi import (
    ArtifactState,
    DocArtifactCreate,
    DocArtifactUpdate,
    ModelArtifactCreate,
    ModelArtifactUpdate,
    DataSetCreate,
    DataSetUpdate,
    MetricCreate,
    MetricUpdate,
    ParameterCreate,
    ParameterUpdate,
    ParameterType,
)
from mr_openapi import (
    DocArtifact as DocArtifactBaseModel,
)
from mr_openapi import (
    ModelArtifact as ModelArtifactBaseModel,
)

from .base import BaseResourceModel

A = TypeVar("A", bound="Artifact")


class Artifact(BaseResourceModel, ABC):
    """Base class for all artifacts.

    Attributes:
        name: Name of the artifact.
        state: State of the artifact.
        description: Description of the artifact.
        external_id: Customizable ID. Has to be unique among instances of the same type.
    """

    name: str | None = None
    state: ArtifactState = ArtifactState.UNKNOWN
    description: str | None = None
    external_id: str | None = None
    experiment_id: str | None = None
    experiment_run_id: str | None = None

    @classmethod
    def from_artifact(cls: type[A], source: ArtifactBaseModel) -> A:
        """Convert a base artifact."""
        model = source.actual_instance
        assert model
        return cls.from_basemodel(model)

    @staticmethod
    def validate_artifact(
        source: ArtifactBaseModel,
    ) -> DocArtifact | ModelArtifact | DataSet | Metric | Parameter:
        """Validate an artifact."""
        model = source.actual_instance
        assert model
        if isinstance(model, DocArtifactBaseModel):
            return DocArtifact.from_basemodel(model)
        if isinstance(model, ModelArtifactBaseModel):
            return ModelArtifact.from_basemodel(model)
        if isinstance(model, DataSetBaseModel):
            return DataSet.from_basemodel(model)
        if isinstance(model, MetricBaseModel):
            return Metric.from_basemodel(model)
        if isinstance(model, ParameterBaseModel):
            return Parameter.from_basemodel(model)
        msg = f"Invalid artifact type: {type(model)}"
        raise ValueError(msg)

    @abstractmethod
    def as_basemodel(self) -> Any:
        """Wrap the object in a BaseModel object."""

    def wrap(self) -> ArtifactBaseModel:
        """Wrap the object in a ArtifactBaseModel object."""
        return ArtifactBaseModel(self.as_basemodel())


class DocArtifact(Artifact):
    """Represents a Document Artifact.

    Attributes:
        name: Name of the document.
        uri: URI of the document.
        description: Description of the object.
        external_id: Customizable ID. Has to be unique among instances of the same type.
    """

    uri: str | None = None

    @override
    def create(self, **kwargs) -> DocArtifactCreate:
        """Create a new DocArtifactCreate object."""
        return DocArtifactCreate(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("id", "custom_properties")),
            artifactType="doc-artifact",
            **kwargs,
        )

    @override
    def update(self, **kwargs) -> DocArtifactUpdate:
        """Create a new DocArtifactUpdate object."""
        return DocArtifactUpdate(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("id", "name", "custom_properties")),
            artifactType="doc-artifact",
            **kwargs,
        )

    @override
    def as_basemodel(self) -> DocArtifactBaseModel:
        return DocArtifactBaseModel(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("custom_properties")),
            artifactType="doc-artifact",
        )

    @classmethod
    @override
    def from_basemodel(cls, source: DocArtifactBaseModel) -> DocArtifact:
        assert source.name
        assert source.state
        return cls(
            id=source.id,
            name=source.name,
            description=source.description,
            external_id=source.external_id,
            create_time_since_epoch=source.create_time_since_epoch,
            last_update_time_since_epoch=source.last_update_time_since_epoch,
            experiment_id=source.experiment_id,
            experiment_run_id=source.experiment_run_id,
            uri=source.uri,
            state=source.state,
            custom_properties=cls._unmap_custom_properties(source.custom_properties)
            if source.custom_properties
            else None,
        )


class ModelArtifact(Artifact):
    """Represents a Model.

    Attributes:
        name: Name of the model.
        uri: URI of the model.
        description: Description of the object.
        external_id: Customizable ID. Has to be unique among instances of the same type.
        model_format_name: Name of the model format.
        model_format_version: Version of the model format.
        storage_key: Storage secret name.
        storage_path: Storage path of the model.
        service_account_name: Name of the service account with storage secret.
        model_source_kind: A string identifier describing the source kind.
        model_source_class: A subgroup within the source kind.
        model_source_group: This identifies a source group for models from source class.
        model_source_id: A unique identifier for a source model within kind, class, and group.
        model_source_name: A human-readable name for the source model.
    """

    # TODO: this could be an enum of valid formats
    model_format_name: str | None = None
    model_format_version: str | None = None
    storage_key: str | None = None
    storage_path: str | None = None
    service_account_name: str | None = None
    model_source_kind: str | None = None
    model_source_class: str | None = None
    model_source_group: str | None = None
    model_source_id: str | None = None
    model_source_name: str | None = None
    uri: str | None = None

    _model_version_id: str | None = None

    @override
    def create(self, **kwargs) -> ModelArtifactCreate:
        """Create a new ModelArtifactCreate object."""
        return ModelArtifactCreate(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("id", "custom_properties")),
            artifactType="model-artifact",
            **kwargs,
        )

    @override
    def update(self, **kwargs) -> ModelArtifactUpdate:
        """Create a new ModelArtifactUpdate object."""
        return ModelArtifactUpdate(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("id", "name", "custom_properties")),
            artifactType="model-artifact",
            **kwargs,
        )

    @override
    def as_basemodel(self) -> ModelArtifactBaseModel:
        return ModelArtifactBaseModel(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("custom_properties")),
            artifactType="model-artifact",
        )

    @classmethod
    @override
    def from_basemodel(cls, source: ModelArtifactBaseModel) -> ModelArtifact:
        """Create a new ModelArtifact object from a BaseModel object."""
        assert source.name
        assert source.uri
        assert source.state
        return cls(
            id=source.id,
            name=source.name,
            description=source.description,
            external_id=source.external_id,
            create_time_since_epoch=source.create_time_since_epoch,
            last_update_time_since_epoch=source.last_update_time_since_epoch,
            experiment_id=source.experiment_id,
            experiment_run_id=source.experiment_run_id,
            uri=source.uri,
            model_format_name=source.model_format_name,
            model_format_version=source.model_format_version,
            storage_key=source.storage_key,
            storage_path=source.storage_path,
            service_account_name=source.service_account_name,
            model_source_kind=source.model_source_kind,
            model_source_class=source.model_source_class,
            model_source_group=source.model_source_group,
            model_source_id=source.model_source_id,
            model_source_name=source.model_source_name,
            state=source.state,
            custom_properties=cls._unmap_custom_properties(source.custom_properties)
            if source.custom_properties
            else None,
        )


class DataSet(Artifact):
    """Represents a DataSet.

    Attributes:
        name: Name of the data set.
        uri: URI of the data set.
        description: Description of the object.
        external_id: Customizable ID. Has to be unique among instances of the same type.
        digest: A unique hash or identifier for the dataset content.
        source_type: The type of source for the dataset.
        source: The location or connection string for the dataset source.
        schema: JSON schema or description of the dataset structure.
        profile: Statistical profile or summary of the dataset.
    """

    uri: str | None = None
    digest: str | None = None
    source_type: str | None = None
    source: str | None = None
    schema: str | None = None  # type: ignore[assignment]
    profile: str | None = None

    @override
    def create(self, **kwargs) -> DataSetCreate:
        """Create a new DataSetCreate object."""
        return DataSetCreate(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("id", "custom_properties")),
            artifactType="dataset-artifact",
            **kwargs,
        )

    @override
    def update(self, **kwargs) -> DataSetUpdate:
        """Create a new DataSetUpdate object."""
        return DataSetUpdate(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("id", "name", "custom_properties")),
            artifactType="dataset-artifact",
            **kwargs,
        )

    @override
    def as_basemodel(self) -> DataSetBaseModel:
        return DataSetBaseModel(
            customProperties=self._map_custom_properties(),
            **self._props_as_dict(exclude=("custom_properties")),
            artifactType="dataset-artifact",
        )

    @classmethod
    @override
    def from_basemodel(cls, source: DataSetBaseModel) -> DataSet:
        """Create a new DataSet object from a BaseModel object."""
        assert source.name
        return cls(
            id=source.id,
            name=source.name,
            description=source.description,
            external_id=source.external_id,
            create_time_since_epoch=source.create_time_since_epoch,
            last_update_time_since_epoch=source.last_update_time_since_epoch,
            experiment_id=source.experiment_id,
            experiment_run_id=source.experiment_run_id,
            uri=source.uri,
            digest=source.digest,
            source_type=source.source_type,
            source=source.source,
            schema=source.var_schema,
            profile=source.profile,
            state=source.state,  # type: ignore[arg-type]
            custom_properties=cls._unmap_custom_properties(source.custom_properties)
            if source.custom_properties
            else None,
        )


class Metric(Artifact):
    """Represents a Metric.

    Attributes:
        name: Name of the metric.
        description: Description of the object.
        external_id: Customizable ID. Has to be unique among instances of the same type.
        value: The numeric value of the metric.
        timestamp: Unix timestamp in milliseconds when the metric was recorded.
        step: The step number for multi-step metrics (e.g., training epochs)
    """

    value: float
    timestamp: str | None = None
    step: int = 0

    @override
    def create(self, **kwargs) -> MetricCreate:
        """Create a new MetricCreate object."""
        return MetricCreate(
            customProperties=self._map_custom_properties(),
            timestamp=self.timestamp,
            **self._props_as_dict(exclude=("id", "timestamp", "custom_properties")),
            artifactType="metric",
            **kwargs,
        )

    @override
    def update(self, **kwargs) -> MetricUpdate:
        """Create a new MetricUpdate object."""
        return MetricUpdate(
            customProperties=self._map_custom_properties(),
            timestamp=self.timestamp,
            **self._props_as_dict(exclude=("id", "name", "timestamp", "custom_properties")),
            artifactType="metric",
            **kwargs,
        )

    @override
    def as_basemodel(self) -> MetricBaseModel:
        return MetricBaseModel(
            customProperties=self._map_custom_properties(),
            timestamp=self.timestamp,
            **self._props_as_dict(exclude=("timestamp", "custom_properties")),
            artifactType="metric",
        )

    @classmethod
    @override
    def from_basemodel(cls, source: MetricBaseModel) -> Metric:
        """Create a new Metric object from a BaseModel object."""
        assert source.name
        return cls(
            id=source.id,
            name=source.name,
            description=source.description,
            external_id=source.external_id,
            create_time_since_epoch=source.create_time_since_epoch,
            last_update_time_since_epoch=source.last_update_time_since_epoch,
            experiment_id=source.experiment_id,
            experiment_run_id=source.experiment_run_id,
            value=source.value,  # type: ignore[arg-type]
            timestamp=source.timestamp,
            step=source.step,  # type: ignore[arg-type]
            state=source.state,  # type: ignore[arg-type]
            custom_properties=cls._unmap_custom_properties(source.custom_properties)
            if source.custom_properties
            else None,
        )


class Parameter(Artifact):
    """Represents a Parameter.

    Attributes:
        name: Name of the parameter.
        description: Description of the object.
        external_id: Customizable ID. Has to be unique among instances of the same type.
        parameter_type: The data type of the parameter (e.g., "string", "number", "boolean", "object").
        value: The value of the parameter.
    """

    value: str | bool | int | float | dict
    parameter_type: ParameterType

    @override
    def create(self, **kwargs) -> ParameterCreate:
        """Create a new ParameterCreate object."""
        return ParameterCreate(
            customProperties=self._map_custom_properties(),
            value=str(self.value),
            **self._props_as_dict(exclude=("id", "value", "custom_properties")),
            artifactType="parameter",
            **kwargs,
        )

    @override
    def update(self, **kwargs) -> ParameterUpdate:
        """Create a new ParameterUpdate object."""
        return ParameterUpdate(
            customProperties=self._map_custom_properties(),
            value=str(self.value),
            **self._props_as_dict(exclude=("id", "name", "value", "custom_properties")),
            artifactType="parameter",
            **kwargs,
        )

    @override
    def as_basemodel(self) -> ParameterBaseModel:
        return ParameterBaseModel(
            customProperties=self._map_custom_properties(),
            value=str(self.value),
            **self._props_as_dict(exclude=("value", "custom_properties")),
            artifactType="parameter",
        )

    @classmethod
    @override
    def from_basemodel(cls, source: ParameterBaseModel) -> Parameter:
        """Create a new Parameter object from a BaseModel object."""
        assert source.name
        assert source.parameter_type
        value = source.value
        if source.parameter_type is ParameterType.NUMBER:
            value = float(value)  # type: ignore[assignment,arg-type]
        elif source.parameter_type is ParameterType.BOOLEAN:
            value = bool(value)  # type: ignore[assignment]
        elif source.parameter_type is ParameterType.OBJECT:
            value = json.loads(value)  # type: ignore[arg-type]
        return cls(
            id=source.id,
            name=source.name,
            description=source.description,
            external_id=source.external_id,
            create_time_since_epoch=source.create_time_since_epoch,
            last_update_time_since_epoch=source.last_update_time_since_epoch,
            experiment_id=source.experiment_id,
            experiment_run_id=source.experiment_run_id,
            value=value,  # type: ignore[arg-type]
            parameter_type=source.parameter_type,
            state=source.state,  # type: ignore[arg-type]
            custom_properties=cls._unmap_custom_properties(source.custom_properties)
            if source.custom_properties
            else None,
        )


ExperimentRunArtifact = Parameter | Metric | DataSet


@dataclass
class ExperimentRunArtifactTypes:
    """Types of experiment run artifacts."""

    params: dict[str, Parameter] = field(default_factory=dict)
    metrics: dict[str, Metric] = field(default_factory=dict)
    datasets: dict[str, DataSet] = field(default_factory=dict)



"""Model Registry REST API.

REST API for Model Registry to create and manage ML model metadata

The version of the OpenAPI document: v1alpha3
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Any, ClassVar

from pydantic import BaseModel, ConfigDict, Field, StrictStr
from typing_extensions import Self

from mr_openapi.models.artifact_state import ArtifactState
from mr_openapi.models.metadata_value import MetadataValue


class ModelArtifactUpdate(BaseModel):
    """An ML model artifact to be updated."""  # noqa: E501

    custom_properties: dict[str, MetadataValue] | None = Field(
        default=None,
        description="User provided custom properties which are not defined by its type.",
        alias="customProperties",
    )
    description: StrictStr | None = Field(default=None, description="An optional description about the resource.")
    external_id: StrictStr | None = Field(
        default=None,
        description="The external id that come from the clients’ system. This field is optional. If set, it must be unique among all resources within a database instance.",
        alias="externalId",
    )
    artifact_type: StrictStr | None = Field(default="model-artifact", alias="artifactType")
    model_format_name: StrictStr | None = Field(
        default=None, description="Name of the model format.", alias="modelFormatName"
    )
    storage_key: StrictStr | None = Field(default=None, description="Storage secret name.", alias="storageKey")
    storage_path: StrictStr | None = Field(
        default=None, description="Path for model in storage provided by `storageKey`.", alias="storagePath"
    )
    model_format_version: StrictStr | None = Field(
        default=None, description="Version of the model format.", alias="modelFormatVersion"
    )
    service_account_name: StrictStr | None = Field(
        default=None, description="Name of the service account with storage secret.", alias="serviceAccountName"
    )
    model_source_kind: StrictStr | None = Field(
        default=None,
        description="A string identifier describing the source kind. It differentiates various sources of model artifacts. This identifier should be agreed upon by producers and consumers of source model metadata. It is not an enumeration to keep the source of model metadata open ended.  E.g. Kubeflow pipelines could use `pipelines` to identify models it produces.",
        alias="modelSourceKind",
    )
    model_source_class: StrictStr | None = Field(
        default=None,
        description="A subgroup within the source kind. It is a specific sub-component or instance within the source kind. E.g. `pipelinerun` for a Kubeflow pipeline run.",
        alias="modelSourceClass",
    )
    model_source_group: StrictStr | None = Field(
        default=None,
        description="Unique identifier for a source group for models from source class.  It maps to a physical group of source models.  E.g. a Kubernetes namespace where the pipeline run was executed.",
        alias="modelSourceGroup",
    )
    model_source_id: StrictStr | None = Field(
        default=None,
        description="A unique identifier for a source model within kind, class, and group. It should be a url friendly string if source supports using URLs to locate source models. E.g. a pipeline run ID.",
        alias="modelSourceId",
    )
    model_source_name: StrictStr | None = Field(
        default=None,
        description="A human-readable name for the source model.  E.g. `my-project/1`, `ibm-granite/granite-3.1-8b-base:2.1.2`.",
        alias="modelSourceName",
    )
    uri: StrictStr | None = Field(
        default=None,
        description="The uniform resource identifier of the physical artifact. May be empty if there is no physical artifact.",
    )
    state: ArtifactState | None = ArtifactState.UNKNOWN
    __properties: ClassVar[list[str]] = [
        "customProperties",
        "description",
        "externalId",
        "artifactType",
        "modelFormatName",
        "storageKey",
        "storagePath",
        "modelFormatVersion",
        "serviceAccountName",
        "modelSourceKind",
        "modelSourceClass",
        "modelSourceGroup",
        "modelSourceId",
        "modelSourceName",
        "uri",
        "state",
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias."""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias."""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Self | None:
        """Create an instance of ModelArtifactUpdate from a JSON string."""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: set[str] = set()

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of each value in custom_properties (dict)
        _field_dict = {}
        if self.custom_properties:
            for _key_custom_properties in self.custom_properties:
                if self.custom_properties[_key_custom_properties]:
                    _field_dict[_key_custom_properties] = self.custom_properties[_key_custom_properties].to_dict()
            _dict["customProperties"] = _field_dict
        return _dict

    @classmethod
    def from_dict(cls, obj: dict[str, Any] | None) -> Self | None:
        """Create an instance of ModelArtifactUpdate from a dict."""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        return cls.model_validate(
            {
                "customProperties": {_k: MetadataValue.from_dict(_v) for _k, _v in obj["customProperties"].items()}
                if obj.get("customProperties") is not None
                else None,
                "description": obj.get("description"),
                "externalId": obj.get("externalId"),
                "artifactType": obj.get("artifactType") if obj.get("artifactType") is not None else "model-artifact",
                "modelFormatName": obj.get("modelFormatName"),
                "storageKey": obj.get("storageKey"),
                "storagePath": obj.get("storagePath"),
                "modelFormatVersion": obj.get("modelFormatVersion"),
                "serviceAccountName": obj.get("serviceAccountName"),
                "modelSourceKind": obj.get("modelSourceKind"),
                "modelSourceClass": obj.get("modelSourceClass"),
                "modelSourceGroup": obj.get("modelSourceGroup"),
                "modelSourceId": obj.get("modelSourceId"),
                "modelSourceName": obj.get("modelSourceName"),
                "uri": obj.get("uri"),
                "state": obj.get("state") if obj.get("state") is not None else ArtifactState.UNKNOWN,
            }
        )
